import numpy as np
from Numeric import pi, sqrt
from tools import factorial
from tools import Rational as Q

"""
This is a script designed for construction of the real solid spherical
harmonics (RSSH) in cartesian form. These can be written as::

         m    |m|  l  |m|
  Y  =  Y  = C    r  P  (cos theta) Phi (phi)
   L     l    l       l                m

where C_l^|m| is a normalization constant
P_l^|m| is the associatied legendre polynomial
and:

              / cos(m phi) , m > 0
  Phi (phi) = |   1        , m = 0
     m        \ sin(-m phi), m < 0

The first few harmonics are listed below::
 +----+---------------------+-__---------------------------------------+
 |  L | l |  m | r^l * Y    | \/ (r^l * Y)                             |
 +----+---s----+------------+------------------------------------------+
 |  0 | 0 |  0 | 1          | (0, 0, 0)                                |
 +----+---p----+------------+------------------------------------------+
 |  1 | 1 | -1 | y          | (0, 1, 0)                                |
 |  2 | 1 |  0 | z          | (0, 0, 1)                                |
 |  3 | 1 |  1 | x          | (1, 0, 0)                                |
 +----+---d----+------------+------------------------------------------+
 |  4 | 2 | -2 | xy         | ( y,  x,  0)                             |
 |  5 | 2 | -1 | yz         | ( 0,  z,  y)                             |
 |  6 | 2 |  0 | 3z^2-r^2   | (-x, -y, 2z)                             |
 |  7 | 2 |  1 | xz         | ( z,  0,  x)                             |
 |  8 | 2 |  2 | x^2-y^2    | ( x, -y,  0)                             |
 +----+---f----+------------+------------------------------------------+
 |  9 | 3 | -3 | 3x^2y-y^3  | (          2xy,       x^2-y^2,        0) |
 | 10 | 3 | -2 | xyz        | (           yz,            xz,       xy) |
 | 11 | 3 | -1 | 5yz^2-yr^2 | (         -2xy, 4z^2-x^2-3y^2,      8yz) |
 | 12 | 3 |  0 | 5z^3-3zr^2 | (         -2xz,          -2yz, 3z^2-r^2) |
 | 13 | 3 |  1 | 5xz^2-xr^2 | (4z^2-3x^2-y^2,          -2xy,      8xz) |
 | 14 | 3 |  2 | x^2z-y^2z  | (          2xz,          -2yz,  x^2-y^2) |
 | 15 | 3 |  3 | x^3-3xy^2  | (      x^2-y^2,          -2xy,        0) |
 +----+--------+----------+--------------------------------------------+

Y_lm is represented as a polynomial in x, y, and z

The function consists of three parts: a normalization constant accessed by
class 'Normalization(l, m)', a polynomial in z accessed with method
'legendre(l, m)', and a polynomial in x and y accessed with method 'Phi(l, m)'

The normalization and the z-polynomial are both invariant of the sign of m
The z-polynomial has powers l-|m|, l-|m|-2, l-|m|-4, l-..., i.e. it is strictly odd (even) if l-|m| is odd (even)
The combined power of x and y is |m| in all terms of Phi
"""

Y_lp = [{}, {}] # Global list of dictionaries for storing calculated
                # Legendre polynomials, and Phi functions

#--------------------------- RELEVANT USER METHODS ---------------------------
def L_to_lm(L):
    """convert L index to (l, m) index"""
    l = int(sqrt(L))
    m = L - l**2 - l
    return l, m

def lm_to_L(l,m):
    """convert (l, m) index to L index"""
    return l**2 + l + m

def Y_to_string(l, m, deriv=None, multiply=None, numeric=False):
    # for L in range(40): print L, Y_to_string(*L_to_lm(L))
    """                                                   l    m
       If deriv is None, return string representation of r  * Y (x, y, z)
                                                               l
       
       If deriv == q, return string is the derivative of above with respect
       to x, y or z if q is 0, 1 or 2 respectively.

       multiply=q indicates that the entire expression should be multiplied by
       x, y or z if q is 0, 1 or 2 respectively.

       numeric=True/False indicates whether the normalization constant should
       be written as a numeric or an algebraic expression.
    """
    assert deriv is None or deriv in range(3)
    assert multiply is None or multiply in range(3)
    
    if deriv is None:
        norm, xyzs = Y_collect(l, m)
    else:
        norm, xyzs = dYdq(l, m, deriv)

    if multiply is not None:
        xyzs = q_times_xyzs(xyzs, multiply)

    string = to_string(l, xyzs, deriv is not None, multiply is not None)
    if string == '0': return '0'
    else: return norm.tostring(numeric) + (' * ' + string) * (string != '1')

def gauss_to_string(l, m, numeric=False):
    """Return string representation of the generalized gaussian::
    
                       _____                           2  
         m            /  1       l!         l+3/2  -a r   l  m
        g (x,y,z) =  / ----- --------- (4 a)      e      r  Y (x,y,z)
         l         \/  4 pi  (2l + 1)!                       l
         
       numeric=True/False indicates whether the normalization constant should 
       be written as a number or an algebraic expression.
    """
    norm, xyzs = Y_collect(l, m)

    ng = Q(2**(2*l+3) * factorial(l), 2 * factorial(2 * l + 1))
    norm.multiply(ng)

    string = to_string(l, xyzs)
    string = (' * ' + string) * (string != '1')
    if numeric:
        snorm = repr(eval(repr(norm.norm)))
    else:
        snorm = repr(norm.norm)
    string = 'sqrt(a**%s*%s)/pi'%(2*l+3, snorm) + string
    string += ' * exp(-a*r2)'

    return string

def gauss_potential_to_string(l, m, numeric=False):
    """Return string representation of the potential of  a generalized
       gaussian.

       The potential is determined by::

          m        m ^    _           m ^
         v [g (r) Y (r) ](r) = v (r) Y (r)
          l  l     l         l     l  l

       where::
               4 pi /  -l-1 /r    l+2         l /oo   1-l      \ 
       v (r) = ---- | r     | dx x   g (r) + r  | dx x   g (r) |
        l      2l+1 \       /0        l         /r        l    /
    """            
    v_l = [[Q(4,1), 1],
           [Q(4,3), 1, 2],
           [Q(4,15), 3, 6, 4],
           [Q(4,105), 15, 30, 20, 8],
           [Q(4,945), 105, 210, 140, 56, 16],
           [Q(4,10395), 945, 1890, 1260, 504, 144, 32],
           ]

    norm, xyzs = Y_collect(l, m)
    norm.multiply(v_l[l][0])

    string = txt_sqrt(norm.norm, numeric) + '*' + (l!=0)*'('
    if numeric:
        string += repr(v_l[l][1] * sqrt(pi))
    else:
        string += str(v_l[l][1]) + '*sqrt(pi)'
    string += '*erf(sqrt(a)*r)'

    if len(v_l[l]) > 2:
        string += '-('
        for n, coeff in enumerate(v_l[l][2:]):
            if n == 0:
                string += str(coeff)
            else:
                string += '+' + str(coeff) + '*(sqrt(a)*r)**%d'%(2*n)
        string += ')*sqrt(a)*r*exp(-a*r2)'

    if l == 0:
        string += '/r'
    elif l == 1:
        string += ')/r/r2*' + to_string(l, xyzs)
    else:
        string += ')/r/r2**%d*'%l + to_string(l, xyzs)

    return string

#----------------------------- TECHNICAL METHODS -----------------------------
def to_string(l, xyzs, deriv=False, multiply=False):
    """Return string representation of an xyz dictionary"""
    if xyzs == {}: return '0'
    out = ''

    for xyz, coef in xyzs.items():
        x, y, z = xyz
        r = l - x - y - z - deriv + multiply
        one = abs(coef) != 1 or (x == 0 and y == 0 and z == 0 and r == 0)
        out += sign(coef) + str(abs(coef)) * one
        out += ('*x'*x + '*y'*y + '*z'*z + '*r2'*(r/2))[1 - one:]

    if out[0] == '+': out = out[1:]
    if len(xyzs) > 1: out = '(' + out + ')'
    return out

def sign(x):
    """Return string representation of the sign of x"""
    if x >= 0: return '+'
    else: return '-'

def txt_sqrt(norm, numeric=False):
    if numeric:
        return repr(sqrt(norm))
    else:
        if sqrt(norm) % 1 == 0:
            return str(sqrt(norm))
        else:
            return 'sqrt(' + str(norm.nom) + \
                   ('./' + str(norm.denom)) * (norm.denom != 1) + ')'

class Normalization:
    """Determine normalization factor of spherical harmonic
                   ______________
             /    / 2l+1   (l-m)!
             |   /  ---- * ------  , m != 0
             | \/   2 pi   (l+m)!
       C  = <      _____
        L    |    / 2l+1
             |   /  ----           , m = 0
             \ \/   4 pi
    """
    def __init__(self, l, m):
        m = abs(m)
        if m == 0:
            self.norm = Q(2 * l + 1, 4)
        else:
            self.norm = Q((2 * l + 1) * factorial(l - m), 2 * factorial(l + m))

    def __str__(self):
        n = self.norm
        sn = sqrt(n)
        if int(sn) == sn:
            string = repr(sn) + '/sqrt(pi)'
        else:
            string = 'sqrt(' + repr(n.nom) + \
                     ('./' + repr(n.denom)) * (n.denom != 1) + '/pi)'
        return string

    def __repr__(self):
        return repr(self.__float__())

    def __float__(self):
        return sqrt(self.norm / pi)

    def multiply(self, x):
        self.norm *= x**2

    def tostring(self, numeric=False):
        if numeric:
            return self.__repr__()
        else:
            return self.__str__()

def legendre(l, m):
    """Determine z dependence of spherical harmonic.
       Returns vector, where the p'th element is the coefficient of
       z^p r^(l-|m|-p).
    """
    # Check if requested has already been calculated
    if (l, m) in Y_lp[0]:
        return Y_lp[0][(l, m)]
    
    m = abs(m)
    assert l >= 0 and 0 <= m <=l
    result = np.zeros(l - m + 1, 'O')
    if l == m == 0:
        """Use that
             0
            P (z) = 1
             0
        """
        result[0] = Q(1)
    elif l == m:
        """Use the recursion relation
            m              m-1
           P (z) = (2m-1) P   (z)
            m              m-1
        """
        result[:] += (2 * m - 1) * legendre(l - 1, m - 1)
    elif l == m + 1:
        """Use the recursion relation
            l-1              l-1
           P  (z) = (2l-1)z P   (z)
            l                l-1
            
        """
        result[1:] += (2 * l - 1) * legendre(l-1, l-1)
    else:
        """Use the recursion relation
            m     2l-1    m       l+m-1  2  m
           P (z)= ---- z P  (z) - ----- r  P  (z)
            l      l-m    l-1      l-m      l-2
        """
        result[1:] += np.multiply(legendre(l - 1, m), Q(2 * l - 1, l - m))
        result[:(l - 2) - m + 1] -= np.multiply(legendre(l - 2, m),
                                                 Q(l + m - 1, l - m))
    # Store result in global dictionary
    Y_lp[0][(l, m)] = result
    return result

def Phi(m):
    """Determine the x and y dependence of the spherical harmonics from
                      |m|   |m|
                   / r   sin  (theta) cos(|m| phi), m >= 0
       Phi (phi) = |
          m        |  |m|   |m|
                   \ r   sin  (theta) sin(|m| phi), m < 0
       Returns dictionary of format {(i, j): c} where c is the coefficient
       of x^i y^j
    """
    # Check if requested has already been calculated
    if m in Y_lp[1]:
        return Y_lp[1][m]
    
    if   m ==  0:
        xys = {(0, 0): 1} # use that Phi_0  = 1
    elif m ==  1:
        xys = {(1, 0): 1} # use that Phi_1  = x
    elif m == -1:
        xys = {(0, 1): 1} # use that Phi_-1 = y
    else:
        """Use the recurrence formula
        
           m > 0:  Phi (x,y) = x Phi   (x,y) - y Phi   (x,y)
                     |m|           |m|-1           1-|m|

           m < 0:  Phi (x,y) = y Phi   (x,y) + x Phi   (x,y)
                     |m|           |m|-1           1-|m|           
        """
        xys  = {}
        phi1 = Phi(abs(m) - 1)
        phi2 = Phi(1 - abs(m))
        for x, y in phi1:
            new = (x + (m > 0), y + (m < 0))
            xys[new] = xys.get(new, 0) +  phi1[(x, y)]
        for x,y in phi2:
            new = (x + (m < 0), y + (m > 0))
            sign = 2 * (m < 0) - 1
            xys[new] = xys.get(new, 0) + sign * phi2[(x, y)]

    # Store result in global dictionary
    Y_lp[1][m] = xys
    return xys

def Y_collect(l, m):
    """Collect all necessary parts of spherical harmonic and return in
       simplified format.
       Return dictionary xyzs has format {(i, j, k): c} where c is the
       coefficient of x^i y^j z^k r^(l-|m|-k), or (since i+j = |m|) the
       coefficient of x^i y^j z^k r^(l-i-j-k), from which it is clear that all
       terms are of power l in x, y and z collectively.
    """
    zs = legendre(l, m)
    xys = Phi(m)

    xyzs = {}
    for xy in xys:
        if xys[xy] != 0:
            for p in range(len(zs)):
                if zs[p] != 0:
                    xyzs[xy + (p,)] = xys[xy] * zs[p]

    # get normalization constant and simplify
    norm = Normalization(l, m)
    norm.multiply(simplify(xyzs))
    
    return norm, xyzs

def Y_collect2(l, m):
    """Same as Y_collect, but collective power of x, y, and z are
    adjusted, such the it is always equal to l (thus avoiding
    multiplication by r)
    """
    norm, p = Y_collect(l, m)
    done = False
    while not done:
        p2 = {}
        done = True
        for (nx, ny, nz), c in p.items():
            n = nx + ny + nz
            if n < l:
                p2[(nx + 2, ny, nz)] = p2.get((nx + 2, ny, nz), 0) + c
                p2[(nx, ny + 2, nz)] = p2.get((nx, ny + 2, nz), 0) + c
                p2[(nx, ny, nz + 2)] = p2.get((nx, ny, nz + 2), 0) + c
                if n + 2 < l:
                    done = False
            else:
                assert n == l
                p2[(nx, ny, nz)] = p2.get((nx, ny, nz), 0) + c
        p = p2
    p2 = p.copy()
    for n, c in p.items():
        if c == 0:
            del p2[n]
    return norm, p2

def dYdq(l, m, q):
    """Returns a normalization constant, and a dictionary describing
       the functional form of the derivative of r^l Y_l^m(x,y,z) with
       respect to x, y or z if q is either 0, 1 or 2 respectively. The
       format of the output dictionary is {(i, j, k): c}, where c is the
       coefficient of x^i y^j z^k r^(l-i-j-k-1).
    """
    norm, xyzs = Y_collect(l, m)
    dxyzs = {}
    
    for xyz, coef in xyzs.items():
        x, y, z = xyz
        r = l - x - y - z

        # chain rule: diff coordinate q only
        if xyz[q] != 0:
            dxyz = list(xyz)
            dxyz[q] -= 1
            dxyz = tuple(dxyz)
            dxyzs[dxyz] = dxyzs.get(dxyz, 0) + xyz[q] * coef
        
        # chain rule: diff coordinate r only
        if r != 0:
            dxyz = list(xyz)
            dxyz[q] += 1
            dxyz = tuple(dxyz)
            dxyzs[dxyz] = dxyzs.get(dxyz, 0) + r * coef

    # remove zeros from list
    for dxyz in dxyzs.keys():
        if dxyzs[dxyz] == 0: dxyzs.pop(dxyz)

    # simplify
    if dxyzs != {}: norm.multiply(simplify(dxyzs))
    return norm, dxyzs

def simplify(xyzs):
    """Rescale coefficients to smallest integer value"""
    norm = Q(1)
    numxyz = np.array(xyzs.values())

    # up-scale all 'xyz' coefficients to integers
    for xyz in numxyz:
        numxyz *= xyz.denom
        norm /= xyz.denom

    # determine least common divisor for 'xyz' coefficients
    dmax = 1
    num_max = max(abs(np.floor(numxyz)))
    for d in range(2, num_max + 1):
        test = numxyz / d
        if np.alltrue(test == np.floor(test)): dmax = d

    # Update simplified dictionary
    norm *= dmax
    for i, xyz in enumerate(xyzs):
        xyzs[xyz] = numxyz[i] / dmax

    return norm

def q_times_xyzs(xyzs, q):
    """multiply xyz dictionary by x, y, or z according to q = 0, 1, or 2"""
    qxyzs = {}
    for xyz, c in xyzs.items():
        qxyz = list(xyz)
        qxyz[q] += 1
        qxyz = tuple(qxyz)
        
        qxyzs[qxyz] = c
    return qxyzs

#--------------------- TEST AND CODE CONSTRUCTING METHODS ---------------------
def orthogonal(L1, L2):
    """Perform the integral
          2pi pi
           /  /
       I = |  |sin(theta) d(theta) d(phi) Y (theta, phi) * Y (theta, phi)
           /  /                            L1               L2
           0  0
       which should be a kronecker delta in L1 and L2
    """
    I = 0.0
    N = 40

    for theta in np.arange(0, pi, pi / N):
        for phi in np.arange(0, 2 * pi, 2 * pi / N):
            x = np.cos(phi) * np.sin(theta)
            y = np.sin(phi) * np.sin(theta)
            z = np.cos(theta)
            r2 = x*x + y*y + z*z
            
            Y1 = eval(Y_to_string(*L_to_lm(L1)))
            Y2 = eval(Y_to_string(*L_to_lm(L2)))

            I += np.sin(theta) * Y1 * Y2
    I *= 2 * (pi / N)**2

    return I

def check_orthogonality(Lmax=10):
    """Check orthogonality for all combinations of the first few harmonics"""
    all_passed = True
    for L1 in range(Lmax+1):
        for L2 in range(L1, Lmax+1):
            I = orthogonal(L1, L2)
            passed =  abs(I - (L1 == L2)) < 3e-3
            all_passed *= passed
            print('L1 = %s,  L2 = %s, passed = %s, I = %s' %(L1, L2, passed, I))
    if all_passed: print('All tests passed')
    else: print('Some tests failed')

def symmetry1(lmax, display=True):
    """Make dictionary of format
       diff = {(l1, m1, q1): (nrel, l2, m2, q2)}
       indicating that
            m1              m2
         d Y             d Y
            l1              l2
         ------ = nrel * ------
          d q1            d q2       
    """
    diff = {} # diff[(l1, m1, q1)] = (nrel, l2, m2, q2)
    unique_L = [] # unique_L[L] = (l, m, q, norm, dxyzs)
    for L in range((lmax + 1)**2):
        l, m = L_to_lm(L)
        for q in range(3):
            identical = False
            name = (l, m, 'xyz'[q])

            norm, dxyzs = dYdq(l, m, q)

            for unique in unique_L:
                if dxyzs == unique[4]:
                    diff[name] = (norm.eval() / unique[3],) + unique[0:3]
                    identical = True
                    break
            if identical == False:
                unique_L.append(name + (norm.eval(), dxyzs))
    if display:
        for key, value in diff.items():
            print(str(key) + ' = ' + str(value[0]) + ' * ' + str(value[1:]))
    else: return diff

def symmetry2(l, display=True):
    """Make dictionary of format
       diff = {(l1, m1, q1): (nrel, l2, m2, q2)}
       indicating that
              m1              m2
           d Y             d Y
              l1              l2
           ------ = nrel * ------
            d q1            d q2
       and
                m1                m2
          q1 * Y   = nrel * q2 * Y
                l1                l2
    """
    diff = {} # diff[(l1, m1, q1)] = (nrel, l2, m2, q2)
    unique_L = [] # unique_L[L] = (l, m, q, dnorm, dxyzs, qnorm, qxyzs)
    for m in range(-l, l+1):
        for q in range(3):
            identical = False
            name = (l, m, q)

            qnorm, xyzs = Y_collect(l, m)
            qxyzs = q_times_xyzs(xyzs, q)
            dnorm, dxyzs = dYdq(l, m, q)
            
            for unique in unique_L:
                if dxyzs == unique[4] and qxyzs == unique[6]:
                    dnrel = dnorm.eval() / unique[3]
                    qnrel = qnorm.eval() / unique[5]
                    print(dnrel == qnrel)
                    if dnrel == qnrel:
                        diff[name] = (dnrel,) + unique[0:3]
                        identical = True
                        break
            if identical == False:
                unique_L.append(name + (dnorm.eval(), dxyzs,
                                        qnorm.eval(), qxyzs))
    if display:
        for key, value in diff.items():
            print(str(key) + ' = ' + str(value[0]) + ' * ' + str(value[1:]))
    else: return diff

def construct_spherical_harmonic_c_function(file, lmax, funcname,
                                            multiply=None, deriv=None):
    """Construct a macro for evaluating values of spherical harmonics,
    or the derivative of any spherical harmonic with respect to some axis.
    The deriv keyword corresponds to that of the Y_to_string function."""
    w = file.write
    indent = 0
    def wn(string=''):
        w(2 * indent * ' ')
        w(string)
        w('\\\n')
    wn('#define %s(l, f, x, y, z, r2, p) (' % funcname)
    indent = 2
    wn('{')
    wn('  switch(l)')
    wn('    {')
    switchindent = 3
    indent += switchindent
    for l in range(lmax + 1):
        wn('case %d:' % l)
        indent += 1
        for M, m in enumerate(range(-l, l + 1)):
            Ystr = Y_to_string(l, m, numeric=True, deriv=deriv)
            wn('p[%d] = f * %s;' % (M, Ystr))
        wn('break;')
        indent -= 1
    wn('default:')
    wn('  assert(0 == 1);')
    indent -= switchindent
    wn('    }')
    wn('}')
    indent = 0
    wn(')')
    w('\n')


def construct_spherical_harmonic_c_code(filename='spherical_harmonics.h',
                                        lmax=4):
    """Construct macros for evaluating spherical harmonics as well as their
    derivatives."""
    file = open(filename, 'w')
    construct = construct_spherical_harmonic_c_function
    construct(file, lmax, 'spherical_harmonics')
    for c in range(3):
        construct(file, lmax, 'spherical_harmonics_derivative_%s' % 'xyz'[c],
                  multiply=c, deriv=c)
    file.close()

    
def construct_c_code(file='temp.c', lmax=3):
    """Method for generating the code in c/spline.c"""
    txt = '//Computer generated code! Hands off!'
    start_func = """
    
// inserts values of f(r) r^l Y_lm(theta, phi) in elements of input array 'a'
void bmgs_radial3(const bmgsspline* spline, int m, 
                  const int n[3], 
                  const double C[3],
                  const double h[3],
                  const double* f, double* a)
{
  int l = spline->l;
  if (l == 0)
    for (int q = 0; q < n[0] * n[1] * n[2]; q++)
      a[q] = 0.28209479177387814 * f[q];
"""
    start_deriv = """

// insert values of
// d( f(r) * r^l Y_l^m )                           d( r^l Y_l^m )
// --------------------- = g(r) q r^l Y_l^m + f(r) --------------
//        dq                                             dq
// where q={x, y, z} and g(r) = 1/r*(df/dr)
void bmgs_radiald3(const bmgsspline* spline, int m, int c, 
                  const int n[3], 
                  const double C[3],
                  const double h[3],
                  const double* f, const double* g, double* a)
{
  int l = spline->l;
"""
    start_case = """
    {
      int q = 0;
      double x = C[0];
      for (int i0 = 0; i0 < n[0]; i0++)
        {
          double y = C[1];
          for (int i1 = 0; i1 < n[1]; i1++)
            {
              double z = C[2];
              for (int i2 = 0; i2 < n[2]; i2++, q++)
                {
"""
    end_case = """
                  z += h[2];
                }
              y += h[1];
            }
          x += h[0];
        }
    }
"""
    
    # insert code for evaluating the function
    txt += start_func
    for l in range(1, lmax + 1):
        txt += '  else if (l == %s)' %l
        txt += start_case
        case = ''
        for m in range(-l, l+1):
            if m == -l: case += ' ' * 18 + 'if (m == %s)\n' %m
            elif m == l: case += '\n' + ' ' * 18 +'else\n'
            else: case += '\n' + ' ' * 18 + 'else if (m == %s)\n' %m
            case += ' ' * 20 + 'a[q] = f[q] * '
            case += Y_to_string(l,m, numeric=True) + ';'
        if 'r2' in case: txt += ' ' * 18 + 'double r2 = x*x+y*y+z*z;\n'
        txt += case
        txt += end_case
    txt += """  else
    assert(0 == 1);
}
"""
    
    # insert code for evaluating the derivative
    txt += start_deriv
    for q in range(3):
        txt += '  // ' + 'xyz'[q] + '\n'
        for l in range(0, lmax + 1):
            if l == 0 and q == 0:
                txt += '  if (c == 0 && l == 0)'
            else: txt += '  else if (c == %s && l == %s)' %(q, l)
        
            txt += start_case
            case = ''
            for m in range(-l, l+1):
                if m == -l: case += ' ' * 18 + 'if (m == %s)\n' %m
                elif m == l: case += '\n' + ' ' * 18 + 'else\n'
                else: case += '\n' + ' ' * 18 + 'else if (m == %s)\n' %m
                case += ' ' * 20 + 'a[q] = g[q] * '
                case += Y_to_string(l, m, multiply=q, numeric=True)
                diff = Y_to_string(l, m, deriv=q, numeric=True)
                if diff != '0':
                    case += ' + f[q] * ' + diff
                case += ';'
            if 'r2' in case: txt += ' ' * 18 + 'double r2 = x*x+y*y+z*z;\n'
            txt += case
            txt += end_case
    txt += """  else
      assert(0 == 1);
}
"""
    f = open(file, 'w')
    print(txt, file=f)
    f.close()
    
def construct_gauss_code(lmax=2):
    """Method for generating the code in gpaw/utilities/gauss.py"""
    Lmax = (lmax + 1)**2
    out= 'Y_L = [\n'
    for L in range(Lmax):
        l, m = L_to_lm(L)
        out+= '  \'' + Y_to_string(l, m, numeric=True) + '\',\n'
    out += ']'

    out += '\ngauss_L = [\n'
    for L in range(Lmax):
        l, m = L_to_lm(L)
        out += '  \'' + gauss_to_string(l, m, numeric=True) + '\',\n'
    out += ']'
    
    out += '\ngausspot_L = [\n'
    for L in range(Lmax):
        l, m = L_to_lm(L)
        out += '  \'' + gauss_potential_to_string(l, m, numeric=True) + '\',\n'
    out += ']'
    
    print(out)

def construct_spherical_code(lmax=3):
    """Method for generating the code in gpaw/spherical_harmonics.py"""
    YL = []
    norms = []
    for L in range((lmax+1)**2):
        #norm, xyzs = Y_collect(*L_to_lm(L))
        norm, xyzs = Y_collect2(*L_to_lm(L))
        norms.append(str(norm))
        YL.append(zip(xyzs.values(), xyzs.keys()))

    print('Y_L = [')
    for L, Y in enumerate(YL):
        l = sqrt(L)
        if l % 1 == 0:
            print('  #' + 'spdfghijklmn'[int(l)] + ':')
        print('  %s,' % Y)
    print(']')
    print('norms =', norms)


if __name__ == '__main__':
    construct_spherical_harmonic_c_code()



from ase.parallel import parprint

from gpaw import GPAW
from gpaw.poisson import PoissonSolver
from gpaw.mpi import world, size

from gpaw.lrtddft2 import LrTDDFT2
from gpaw.lrtddft2.lr_communicators import LrCommunicators

# atoms, gs_calc = restart('r-methyl-oxirane.gpw')

dd_size = 2 * 2 * 2
eh_size = size // dd_size
assert eh_size * dd_size == size

lr_comms = LrCommunicators(world, dd_size, eh_size)

gs_calc = GPAW('r-methyl-oxirane.gpw',
               communicator=lr_comms.dd_comm,
               txt='lr_gs.txt',
               poissonsolver=PoissonSolver(eps=1e-20,
                                           remove_moment=(1 + 3 + 5)))

lr = LrTDDFT2('lrtddft2',
              gs_calc,
              fxc='LDA',
              min_occ=0,            # usually zero
              max_occ=15,           # a few above LUMO
              min_unocc=10,         # a few before HOMO
              max_unocc=20,         # number of converged states
              max_energy_diff=8.0,  # eV
              recalculate=None,
              lr_communicators=lr_comms,
              txt='-')

# This is the expensive part!
lr.calculate()

# transitions
(w, S, R, Sx, Sy, Sz) = lr.get_transitions('trans.dat')
# spectrum
lr.get_spectrum('spectrum.dat', 0, 10.)


parprint('')
parprint(' %5s => %5s  contribution' % ('occ', 'unocc'))
f2 = lr.lr_transitions.get_transition_contributions(0)
for (ip, val) in enumerate(f2):
    if (val > 1e-3):
        parprint(' %5d => %5d  %lf %%\n ' %
                 (lr.ks_singles.kss_list[ip].occ_ind,
                  lr.ks_singles.kss_list[ip].unocc_ind,
                  val / 2. * 100))



from ase import Atoms
from gpaw import GPAW, PW
h = Atoms('H', cell=(5, 5, 5))
h.center()
h.calc = GPAW(setups='ae', txt='H.ae.txt')
for ecut in range(200, 1001, 100):
    h.calc.set(mode=PW(ecut))
    e = h.get_potential_energy()



import numpy as np
from math import sqrt
from ase import Atoms
from ase.build import hcp0001
from gpaw import GPAW

# Vacuum and hcp lattice parameter for graphene
d = 4.0
a = 2.4437

calc = GPAW(h=0.15, xc='LDA', nbands=-4, txt='-',
            basis='dzp', convergence={'energy': 1e-5, 'density': 1e-5})

# Calculate potential energy per atom for orthogonal unitcell
atoms = hcp0001('C', a=a / sqrt(3), vacuum=d, size=(3, 2, 1), orthogonal=True)
del atoms[[1, -1]]
atoms.center(axis=0)
atoms.calc = calc

kpts_c = np.ceil(50 / np.sum(atoms.get_cell()**2, axis=1)**0.5).astype(int)
kpts_c[~atoms.get_pbc()] = 1
calc.set(kpts=kpts_c)
eppa1 = atoms.get_potential_energy() / len(atoms)
F1_av = atoms.get_forces()
assert np.abs(F1_av).max() < 5e-3

# Redo calculation with non-orthogonal unitcell
atoms = Atoms(symbols='C2', pbc=(True, True, False),
              positions=[(a / 2, -sqrt(3) / 6 * a, d),
                         (a / 2, sqrt(3) / 6 * a, d)],
              cell=[(a / 2, -sqrt(3) / 2 * a, 0),
                    (a / 2, sqrt(3) / 2 * a, 0),
                    (0, 0, 2 * d)])
atoms.calc = calc

kpts_c = np.ceil(50 / np.sum(atoms.get_cell()**2, axis=1)**0.5).astype(int)
kpts_c[~atoms.get_pbc()] = 1
calc.set(kpts=kpts_c)
eppa2 = atoms.get_potential_energy() / len(atoms)
F2_av = atoms.get_forces()
assert np.abs(F2_av).max() < 5e-3
assert abs(eppa1 - eppa2) < 1e-3



from ase.build import fcc100, add_adsorbate
from gpaw import GPAW

slab = fcc100('Al', (2, 2, 2), a=4.05, vacuum=7.5)
add_adsorbate(slab, 'Na', 4.0)
slab.center(axis=2)

slab.calc = GPAW(txt='zero.txt',
                 xc='PBE',
                 setups={'Na': '1'},
                 kpts=(4, 4, 1))
e1 = slab.get_potential_energy()
slab.calc.write('zero.gpw')

slab.pbc = True
slab.calc.set(txt='periodic.txt')
e2 = slab.get_potential_energy()
slab.calc.write('periodic.gpw')

slab.pbc = (True, True, False)
slab.calc.set(poissonsolver={'dipolelayer': 'xy'},
              txt='corrected.txt')
e3 = slab.get_potential_energy()
slab.calc.write('corrected.gpw')



from ase import Atoms
from gpaw import GPAW
import numpy as np


def test_atoms_too_close():
    atoms = Atoms('H2', [(0.0, 0.0, 0.0),
                         (0.0, 0.0, 3.995)],
                  cell=(4, 4, 4), pbc=True)

    calc = GPAW(txt=None)
    atoms.calc = calc
    try:
        calc.initialize(atoms)
        calc.set_positions(atoms)
    except RuntimeError as err:
        print('got error as expected: {}'.format(err))
    else:
        # silly exception where we actually skip the check for older numpies
        if hasattr(np, 'divmod'):
            assert 2 + 2 == 5



from ase.build import mx2

from gpaw import GPAW, FermiDirac

atoms = mx2(formula='HfS2', kind='1T', a=3.648, thickness=2.895,
            size=(1, 1, 1), vacuum=12.0)
atoms.center(vacuum=6.0, axis=2)
name = atoms.get_chemical_formula(mode='hill')

h = 0.18
kx = 9
ky = 9
kz = 1

calc = GPAW(mode='lcao',
            h=h,
            kpts={'size': (kx, ky, kz), 'gamma': True},
            xc='PBE',
            basis='dzp',
            parallel={'band': 1},
            symmetry='off',
            convergence={'bands': -2},
            maxiter=600,
            txt=None,
            occupations=FermiDirac(width=0.01))

atoms.calc = calc
atoms.get_potential_energy()
calc.write(name + '.gpw')



from myqueue.task import task


def create_tasks():
    return [task('h2_gs.py'),
            task('h2_diss.py@8:10m', deps='h2_gs.py'),
            task('graphene_h_gs.py@8:10m'),
            task('graphene_h_prop.py@32:2h', deps='graphene_h_gs.py')]



# creates: occupation_numbers.png
import numpy as np
import matplotlib.pyplot as plt

from gpaw.occupations import fermi_dirac, marzari_vanderbilt, methfessel_paxton

width = 0.05
x = np.linspace(-0.2, 0.2, 101)

ax = plt.subplot()
ax.plot(x, fermi_dirac(x, 0.0, width)[0],
        label='fermi-dirac')
ax.plot(x, marzari_vanderbilt(x, 0.0, width)[0],
        label='marzari-vanderbilt')
ax.plot(x, methfessel_paxton(x, 0.0, width, 0)[0],
        label='methfessel_paxton-0')
ax.plot(x, methfessel_paxton(x, 0.0, width, 1)[0],
        label='methfessel_paxton-1')
plt.xlabel('energy [eV]')
plt.ylabel('occupation')
plt.legend()
# plt.show()
plt.savefig('occupation_numbers.png')



import sys
import sphinx_rtd_theme
from gpaw import __version__

assert sys.version_info >= (3, 6)

sys.path.append('.')

extensions = ['images',
              'ext',
              'sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.extlinks',
              'sphinx.ext.viewcode',
              'sphinx.ext.mathjax',
              'sphinx.ext.intersphinx']
extlinks = {'doi': ('https://doi.org/%s', 'doi:'),
            'arxiv': ('https://arxiv.org/abs/%s', 'arXiv:')}
templates_path = ['templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'GPAW'
copyright = '2020, GPAW developers'
exclude_patterns = ['build']
default_role = 'math'
pygments_style = 'sphinx'
autoclass_content = 'both'
modindex_common_prefix = ['gpaw.']
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.8', None),
    'ase': ('https://wiki.fysik.dtu.dk/ase', None),
    'numpy': ('https://docs.scipy.org/doc/numpy', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'pytest': ('https://docs.pytest.org/en/stable', None),
    'mayavi': ('http://docs.enthought.com/mayavi/mayavi', None)}
nitpick_ignore = [('py:class', 'gpaw.calculator.GPAW'),
                  ('py:class', 'gpaw.spinorbit.BZWaveFunctions'),
                  ('py:class', 'GPAW'),
                  ('py:class', 'ase.spectrum.dosdata.GridDOSData'),
                  ('py:class', 'ase.atoms.Atoms'),
                  ('py:class', 'gpaw.point_groups.group.PointGroup')]

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_style = 'gpaw.css'
html_title = 'GPAW'
html_favicon = 'static/gpaw_favicon.ico'
html_static_path = ['static']
html_last_updated_fmt = '%a, %d %b %Y %H:%M:%S'
dev_version = '21.1.1b1'  # This line auto-edited by newrelease script
stable_version = '21.1.0'  # This line auto-edited by newrelease script
html_context = {
    'current_version': __version__,
    'versions':
        [(f'{dev_version} (development)',
          'https://wiki.fysik.dtu.dk/gpaw/dev'),
         (f'{stable_version} (latest stable)',
          'https://wiki.fysik.dtu.dk/gpaw')]}
mathjax_config = {
    'TeX': {
        'Macros': {
            'br': '{\\mathbf r}',
            'bk': '{\\mathbf k}',
            'bG': '{\\mathbf G}'}}}



import os
import shutil
from pathlib import Path
from myqueue.task import task


def create_tasks():
    if os.getenv('AGTS_FILES'):
        dir = Path(os.getenv('AGTS_FILES'))
        for file in [Path('lifepo4_wo_li.traj'),
                     Path('NEB_init.traj')]:
            if not file.is_file():
                shutil.copyfile(dir / file, file)

    t1 = task('batteries1.py', tmax='1h')
    t2 = task('batteries2.py', tmax='3h')
    t3 = task('batteries3.py', tmax='1h', cores=8, deps=[t1, t2])
    return [t1, t2, t3]



from gpaw.solvation.sjm import SJM, SJMPower12Potential

from ase.build import fcc111
from gpaw import FermiDirac

# Import solvation modules
from ase.data.vdw import vdw_radii
from gpaw.solvation import (
    EffectivePotentialCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction)

# Solvent parameters
u0 = 0.180  # eV
epsinf = 78.36  # Dielectric constant of water at 298 K
gamma = 0.00114843767916  # 18.4*1e-3 * Pascal* m
T = 298.15   # K


def atomic_radii(atoms):
    return [vdw_radii[n] for n in atoms.numbers]


# Structure is created
atoms = fcc111('Au', size=(1, 1, 3))
atoms.center(axis=2, vacuum=10)
atoms.translate([0, 0, -2])

# SJM parameters
potential = 3.4
ne = 0.1
dpot = 0.01

# The calculator
calc = SJM(doublelayer={'upper_limit': 23},
           potential=potential,
           dpot=dpot,
           ne=ne,
           verbose=True,

           gpts=(16, 16, 136),
           poissonsolver={'dipolelayer': 'xy'},
           kpts=(9, 9, 1),
           xc='PBE',
           txt=f'Au_pot_{potential:1.2f}.txt',
           occupations=FermiDirac(0.1),
           cavity=EffectivePotentialCavity(
               effective_potential=SJMPower12Potential(atomic_radii, u0),
               temperature=T,
               surface_calculator=GradientSurface()),
           dielectric=LinearDielectric(epsinf=epsinf),
           interactions=[SurfaceInteraction(surface_tension=gamma)])

# Run the calculation
atoms.calc = calc
atoms.get_potential_energy()



# Creates: graphene_eps.png
import matplotlib.pyplot as plt
import numpy as np
from ase import Atoms
from ase.units import Hartree, Bohr
from ase.utils import seterr

from gpaw import GPAW, FermiDirac
from gpaw.wavefunctions.pw import PW
from gpaw.response.df import DielectricFunction
from gpaw.mpi import world
from gpaw.bztools import find_high_symmetry_monkhorst_pack

a = 2.5
c = 3.22

pbc = [True, True, False]

GR = Atoms(symbols='C2', positions=[(0.5 * a, -np.sqrt(3) / 6 * a, 0.0),
                                    (0.5 * a, np.sqrt(3) / 6 * a, 0.0)],
           cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                 (0.5 * a, 0.5 * 3**0.5 * a, 0),
                 (0.0, 0.0, c * 2.0)], pbc=True)
old_cell = GR.get_cell()
old_cell[2, 2] = 2 * c
GR.set_cell(old_cell)
atoms = GR

calc = GPAW(h=0.18,
            mode=PW(400),
            kpts={'density': 10.0, 'gamma': True},
            occupations=FermiDirac(0.1))

atoms.calc = calc
atoms.get_potential_energy()
calc.write('gs.gpw')

kpts = find_high_symmetry_monkhorst_pack('gs.gpw', density=30,
                                         pbc=pbc)
responseGS = GPAW('gs.gpw').fixed_density(
    kpts=kpts,
    parallel={'band': 1},
    nbands=30,
    occupations=FermiDirac(0.001),
    convergence={'bands': 20})

responseGS.get_potential_energy()
responseGS.write('gsresponse.gpw', 'all')

df = DielectricFunction('gsresponse.gpw', eta=25e-3,
                        pbc=pbc, domega0=0.01,
                        integrationmode='tetrahedron integration')
df1tetra, df2tetra = df.get_dielectric_function(q_c=[0, 0, 0])

df = DielectricFunction('gsresponse.gpw',
                        domega0=0.01,
                        eta=25e-3)
df1, df2 = df.get_dielectric_function(q_c=[0, 0, 0])
omega_w = df.get_frequencies()

if world.rank == 0:
    plt.figure(figsize=(6, 6))
    plt.plot(omega_w, df2.imag * 2, label='Point sampling')
    plt.plot(omega_w, df2tetra.imag * 2, label='Tetrahedron')
    # Analytical result for graphene
    sigmainter = 1 / 4.  # The surface conductivity of graphene
    with seterr(divide='ignore', invalid='ignore'):
        dfanalytic = 1 + (4 * np.pi * 1j / (omega_w / Hartree) *
                          sigmainter / (c / Bohr))

    plt.plot(omega_w, dfanalytic.imag, label='Analytic')

    plt.xlabel('Frequency (eV)')
    plt.ylabel('$\\mathrm{Im}\\varepsilon$')
    plt.xlim(0, 6)
    plt.ylim(0, 50)
    plt.legend()
    plt.tight_layout()
    plt.savefig('graphene_eps.png', dpi=600)



import pickle
import numpy as np
from gpaw import GPAW, FermiDirac
from gpaw.response.g0w0 import G0W0
from ase.build import bulk
from gpaw.wavefunctions.pw import PW

atoms = bulk('BN', 'zincblende', a=3.615)

calc = GPAW(mode=PW(300),
            parallel={'domain': 1},
            kpts={'size': (3, 3, 3), 'gamma': True},
            xc='LDA',
            occupations=FermiDirac(0.001),
            txt='BN_groundstate.txt')

atoms.calc = calc
atoms.get_potential_energy()

calc.diagonalize_full_hamiltonian()
calc.write('BN_gs_fulldiag.gpw', 'all')

gw = G0W0('BN_gs_fulldiag.gpw',
          bands=(1, 7),
          filename='BN_GW0',
          method='GW0',
          maxiter=5,
          mixing=0.5,
          ecut=50)

gw.calculate()

result = pickle.load(open('BN_GW0_results.pckl', 'rb'))

for i in range(result['iqp'].shape[0]):
    print('Ite:', i,
          'Gap:',
          np.min(result['iqp'][i, 0, :, 3]) -
          np.max(result['iqp'][i, 0, :, 2]))



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.wavefunctions.pw import PW


# Plane wave cutoff
pwcutoff = 400.0

# NxNxN k-point sampling
k = 4

# Si lattice constant
alat = 5.421

# bulk calculation
bulk_crystal = bulk('Si', 'diamond', a=alat)
bulk_calc = GPAW(mode=PW(pwcutoff),
                 parallel={'domain': 1, 'band': 1},
                 kpts={'size': (k, k, k), 'gamma': True},  # gamma-centred grid
                 xc='PBE',
                 occupations=FermiDirac(0.01),
                 txt='si.rpa.pbe_output.txt')

bulk_crystal.calc = bulk_calc
e0_bulk_pbe = bulk_crystal.get_potential_energy()

# Now we have the density, but only the occupied states;
# at this point we diagonalise the full Hamiltonian to get
# the empty states as well (expensive)
bulk_calc.diagonalize_full_hamiltonian(nbands=200)

# the 'all' ensures we write wavefunctions too
bulk_calc.write('bulk.all.gpw', mode='all')



from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.wfwriter import WaveFunctionWriter

# Read the ground-state file
td_calc = LCAOTDDFT('gs.gpw', txt='td.out')

# Attach any data recording or analysis tools
DipoleMomentWriter(td_calc, 'dm.dat')
WaveFunctionWriter(td_calc, 'wf.ulm')

# Kick and propagate
td_calc.absorption_kick([1e-5, 0., 0.])
td_calc.propagate(20, 1500)

# Save the state for restarting later
td_calc.write('td.gpw', mode='all')



from numpy.polynomial import Polynomial
import ase.units as u
from ase.build import bulk
from gpaw import GPAW, PW

afcc = 3.985             # Theoretical fcc lattice parameter
a = afcc * 2**(-1 / 3)   # Assuming the same volume per atom
a = afcc * (2 / 3)**0.5  # Assuming the same nearest neighbor distance

bcc = bulk('Al', 'bcc', a=a)
bcc.calc = GPAW()

# Convergence with respect to k-points:
bcc.calc.set(mode=PW(300), txt='Al-bcc-k.txt')

for k in [4, 6, 8, 10]:
    bcc.calc.set(kpts=(k, k, k))
    print(k, bcc.get_potential_energy())

# Convergence with respect to grid spacing:
bcc.calc.set(kpts=(8, 8, 8), txt='Al-bcc-ecut.txt')

for ecut in [200, 300, 400, 500]:
    bcc.calc.set(mode=PW(ecut))
    print(ecut, bcc.get_potential_energy())

# Set parameters to reasonably converged values
E = []
A = [3.0, 3.1, 3.2, 3.3]
for a in A:
    bcc = bulk('Al', 'bcc', a=a)
    bcc.calc = GPAW(mode=PW(300),
                    kpts=(8, 8, 8),
                    txt=f'bulk-bcc-a{a:.1f}.txt')
    E.append(bcc.get_potential_energy())

p = Polynomial.fit(A, E, 3)
a0 = p.deriv(1).roots()[0]
B = p.deriv(2)(a0) * 2 / 9 / a0 / u.J * u.m**3 * 1e-9  # GPa
print(a0, B)

assert abs(a0 - 3.1969) < 0.001
assert abs(B - 77.5) < 0.1



# Test that the atomic corrections of LCAO work correctly,
# by verifying that the different implementations yield the same numbers.
#
# For example the corrections P^* dH P to the Hamiltonian.
#
# This is done by invoking GPAW once for each type of calculation.

from itertools import count

import pytest
from ase.build import molecule, bulk

from gpaw.utilities import compiled_with_sl
from gpaw import GPAW, LCAO
from gpaw.mpi import world


pytestmark = pytest.mark.skipif(not compiled_with_sl(),
                                reason='not compiled_with_sl()')


def run(system, **kwargs):
    corrections = ['dense', 'sparse']

    counter = count()
    energies = []
    for correction in corrections:
        parallel = {}
        if world.size >= 4:
            parallel['band'] = 2
            # if correction.name != 'dense':
            parallel['sl_auto'] = True
        calc = GPAW(mode=LCAO(atomic_correction=correction),
                    basis='sz(dzp)',
                    # spinpol=True,
                    parallel=parallel,
                    txt='gpaw.{}.txt'.format(next(counter)),
                    h=0.35, **kwargs)

        def stopcalc():
            calc.scf.converged = True

        calc.attach(stopcalc, 2)
        system.calc = calc
        energy = system.get_potential_energy()
        energies.append(energy)
        if calc.world.rank == 0:
            print('e', energy)

    master = calc.wfs.world.rank == 0
    if master:
        print('energies', energies)

    eref = energies[0]
    errs = []
    for energy, c in zip(energies, corrections):
        err = abs(energy - eref)
        errs.append(err)
        if master:
            print('err=%e :: name=%s' % (err, correction))

    maxerr = max(errs)
    assert maxerr < 1e-11, maxerr


def test_lcao_atomic_corrections(in_tmp_dir):
    # Use a cell large enough that some overlaps are zero.
    # Thus the matrices will have at least some sparsity.
    system = molecule('CH3CH2OH')
    system.center(vacuum=3.0)
    system.pbc = (0, 1, 1)
    system = system.repeat((1, 1, 2))
    system.rattle(stdev=0.05)

    run(system)

    system2 = bulk('Cu', orthorhombic=True) * (2, 1, 2)
    run(system2, kpts=[2, 3, 4])



import numpy as np
from ase.build import molecule
from gpaw import GPAW, setup_paths
setup_paths.insert(0, '.')

atoms = molecule('H2O')

h = 0.2

for L in np.arange(4, 14, 2) * 8 * h:
    atoms.set_cell((L, L, L))
    atoms.center()
    calc = GPAW(xc='PBE',
                h=h,
                nbands=-40,
                eigensolver='cg',
                setups={'O': 'hch1s'})
    atoms.calc = calc
    e1 = atoms.get_potential_energy()
    calc.write(f'h2o_hch_{L:.1f}.gpw')



def create_tasks():
    from myqueue.task import task
    return [
        task('dos.agts.py',
             deps=['../iron/PBE.py', '../wavefunctions/CO.py',
                   '../wannier/si.py', '../aluminium/Al_fcc.py'])]


if __name__ == '__main__':
    import os
    import sys
    for filename in ['aluminium/Al-fcc.gpw',
                     'wannier/si.gpw',
                     'wavefunctions/CO.gpw',
                     'iron/ferro.gpw',
                     'iron/anti.gpw',
                     'iron/non.gpw']:
        sys.argv = ['', '../' + filename]
        exec(open('dos.py').read())
    if not os.path.islink('ferro.gpw'):
        os.symlink('../iron/ferro.gpw', 'ferro.gpw')
    exec(open('pdos.py').read())



from myqueue.task import task


def create_tasks():
    return [
        task('gs_3x3_defect.py@16:5m'),
        task('unfold_3x3_defect.py@16:10m', deps='gs_3x3_defect.py'),
        task('plot_sf.py', deps='unfold_3x3_defect.py')]



from ase.dft.kpoints import ibz_points, bandpath
from ase.parallel import paropen
from gpaw import GPAW

layer = GPAW('Fe_gs.gpw', txt=None).atoms

points = ibz_points['bcc']
G = points['Gamma']
H = points['H']
P = points['P']
N = points['N']
H_z = [H[0], -H[1], -H[2]]
G_yz = [2 * H[0], 0.0, 0.0]

kpts, x, X = bandpath([G, H, G_yz], layer.cell, npoints=1000)
calc = GPAW('Fe_gs.gpw').fixed_density(
    kpts=kpts,
    symmetry='off',
    txt='Fe_bands.txt',
    parallel={'band': 1})

calc.write('Fe_bands.gpw')

f = paropen('Fe_kpath.dat', 'w')
for k in x:
    print(k, file=f)
f.close()

f = paropen('Fe_highsym.dat', 'w')
for k in X:
    print(k, file=f)
f.close()



compiler = 'gcc'
mpicompiler = '/opt/SUNWhpc/HPC8.2/gnu/bin/mpicc'
mpilinker = 'MPICH_CC=gcc mpicc -Xlinker --export-dynamic'

extra_compile_args = ['-O3', '-std=c99', '-fpic']

scalapack = False

mkl_dir = '/opt/gridware/intel/Compiler/11.1/056/mkl/lib/em64t/'

library_dirs += [mkl_dir]
libraries = ['mkl_intel_lp64' ,'mkl_sequential' ,'mkl_core',
             'mkl_lapack',
             ]

extra_link_args = ['-Wl,-rpath=' + mkl_dir]




from ase import Atoms
from ase.optimize import BFGS

from gpaw import GPAW
from gpaw.wavefunctions.pw import PW
from gpaw.mpi import world


def test_pw_slab():
    a = 2.65
    slab = Atoms('Li2',
                 [(0, 0, 0), (0, 0, a)],
                 cell=(a, a, 3 * a),
                 pbc=True)
    k = 4
    calc = GPAW(mode=PW(200),
                eigensolver='rmm-diis',
                parallel={'band': min(world.size, 4)},
                idiotproof=0,
                kpts=(k, k, 1))
    slab.calc = calc
    BFGS(slab).run(fmax=0.01)
    assert abs(slab.get_distance(0, 1) - 2.46) < 0.01



import pytest
import numpy as np
from ase.build import molecule

from gpaw import GPAW
from gpaw.tddft import TDDFT
from gpaw.poisson import PoissonSolver
from gpaw.mpi import world
from gpaw.test import equal


@pytest.mark.gllb
@pytest.mark.libxc
def test_tddft_restart(in_tmp_dir):
    atoms = molecule('SiH4')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=7, h=0.4,
                poissonsolver=PoissonSolver(eps=1e-16),
                convergence={'density': 1e-8},
                xc='GLLBSC',
                txt='gs.out')
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = TDDFT('gs.gpw', txt='td.out')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3, 'dm.dat')

    # Write a restart point
    td_calc.write('td.gpw', mode='all')

    # Keep propagating
    td_calc.propagate(20, 3, 'dm.dat')

    # Restart from the restart point
    td_calc = TDDFT('td.gpw', txt='td2.out')
    td_calc.propagate(20, 3, 'dm.dat')
    world.barrier()

    # Check dipole moment file
    data_tj = np.loadtxt('dm.dat')
    # Original run
    ref_i = data_tj[4:6].ravel()
    # Restarted steps
    data_i = data_tj[7:].ravel()

    tol = 1e-10
    equal(data_i, ref_i, tol)

    # Test the absolute values
    data = np.loadtxt('dm.dat')[:6].ravel()
    if 0:
        from gpaw.test import print_reference
        print_reference(data, 'ref', '%.12le')

    ref = [0.0, 1.46915686e-15, -3.289312570937e-14, -2.273046460905e-14,
           -3.201827522804e-15, 0.82682747, 1.30253044e-15, 6.113786782415e-05,
           6.113753835177e-05, 6.113827464597e-05, 1.65365493, -1.69353262e-15,
           0.0001073089135539, 0.0001073052457949,
           0.0001073068465827, 2.4804824,
           1.42934356e-15, 0.0001353894007493, 0.0001353887214486,
           0.0001353873226291, 3.30730987, -3.43926271e-16, 0.0001441529062519,
           0.000144155244532, 0.0001441536382364, 4.13413733, -8.41062896e-16,
           0.0001348222114923, 0.0001348264801341, 0.0001348229035149]

    print(data.tolist())

    tol = 1e-7
    equal(data, ref, tol)



import pytest
import numpy as np
from gpaw import GPAW, PW
from ase import Atoms
from gpaw.nlopt.shift import get_shift
from gpaw.nlopt.matrixel import make_nlodata
from gpaw.mpi import world


@pytest.mark.skipif(world.size > 4, reason='System too small')
def test_shift(in_tmp_dir):
    # Check for Hydrogen atom
    atoms = Atoms('H', cell=(3 * np.eye(3)), pbc=True)

    # Do a GS and save it
    calc = GPAW(
        mode=PW(600), symmetry={'point_group': False},
        kpts={'size': (2, 2, 2)}, nbands=5, txt=None)
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', 'all')

    # Get the mml
    make_nlodata()

    # Do a shift current caclulation
    get_shift(freqs=np.linspace(0, 5, 100))

    # Check it
    if world.rank == 0:
        shift = np.load('shift.npy')

        # Check for nan's
        assert not np.isnan(shift).any()
        # It should be zero (small) since H2 is centro-symm.
        assert np.all(np.abs(shift[1]) < 1e-8)



from time import time
import numpy as np
import ase.db
from ase.test.tasks.dcdft import DeltaCodesDFTCollection as Collection
from gpaw import GPAW, PW, FermiDirac

c = ase.db.connect('dcdft.db')

ecut = 340
kptdensity = 3.5
width = 0.10

collection = Collection()

for name in ['K', 'Ca', 'Ti']:
    atoms = collection[name]
    cell = atoms.get_cell()
    # Loop over volumes:
    for n, x in enumerate(np.linspace(0.98, 1.02, 5)):
        id = c.reserve(name=name, x=x)
        if id is None:
            # This calculation has been or is being done:
            continue

        atoms.set_cell(cell * x, scale_atoms=True)
        atoms.calc = GPAW(txt='%s-%d.txt' % (name, n),
                          mode=PW(ecut),
                          xc='PBE',
                          kpts={'density': kptdensity},
                          parallel={'band': 1},
                          idiotproof=False,
                          occupations=FermiDirac(width))

        t1 = time()
        atoms.get_potential_energy()
        t2 = time()

        # Write to database:
        c.write(atoms, name=name, x=x, time=t2 - t1,
                ecut=ecut, kptdensity=kptdensity, width=width)

        del c[id]



from pathlib import Path

import numpy as np
from scipy.spatial import cKDTree

from ase.utils import convert_string_to_fd
from ase.utils.timing import Timer, timer

from gpaw import GPAW
import gpaw.mpi as mpi
from gpaw.response.math_func import two_phi_planewave_integrals


class KohnShamKPoint:
    """Kohn-Sham orbital information for a given k-point."""
    def __init__(self, n_t, s_t, K,
                 eps_t, f_t, ut_tR, projections, shift_c):
        """K-point data is indexed by a joint spin and band index h, which is
        directly related to the transition index t."""
        self.K = K                      # BZ k-point index
        self.n_t = n_t                  # Band index
        self.s_t = s_t                  # Spin index
        self.eps_t = eps_t              # Eigenvalues
        self.f_t = f_t                  # Occupation numbers
        self.ut_tR = ut_tR              # Periodic part of wave functions
        self.projections = projections  # PAW projections

        self.shift_c = shift_c  # long story - see the
        # PairDensity.construct_symmetry_operators() method


class KohnShamKPointPair:
    """Object containing all transitions between Kohn-Sham orbitals from a
    specified k-point to another."""

    def __init__(self, kpt1, kpt2, mynt, nt, ta, tb, comm=None):
        self.kpt1 = kpt1
        self.kpt2 = kpt2

        self.mynt = mynt  # Number of transitions in this block
        self.nt = nt      # Total number of transitions between all blocks
        self.ta = ta      # First transition index of this block
        self.tb = tb      # First transition index of this block not included
        self.comm = comm  # MPI communicator between blocks of transitions

    def transition_distribution(self):
        """Get the distribution of transitions."""
        return self.mynt, self.nt, self.ta, self.tb

    def get_transitions(self):
        return self.n1_t, self.n2_t, self.s1_t, self.s2_t

    def get_all(self, A_mytx):
        """Get a certain data array with all transitions"""
        if self.comm is None or A_mytx is None:
            return A_mytx

        A_tx = np.empty((self.mynt * self.comm.size,) + A_mytx.shape[1:],
                        dtype=A_mytx.dtype)

        self.comm.all_gather(A_mytx, A_tx)

        return A_tx[:self.nt]

    @property
    def n1_t(self):
        return self.get_all(self.kpt1.n_t)

    @property
    def n2_t(self):
        return self.get_all(self.kpt2.n_t)

    @property
    def s1_t(self):
        return self.get_all(self.kpt1.s_t)

    @property
    def s2_t(self):
        return self.get_all(self.kpt2.s_t)

    @property
    def deps_t(self):
        return self.get_all(self.kpt2.eps_t) - self.get_all(self.kpt1.eps_t)

    @property
    def df_t(self):
        return self.get_all(self.kpt2.f_t) - self.get_all(self.kpt1.f_t)

    @classmethod
    def add_mytransitions_array(cls, _key, key):
        """Add a A_tx data array class attribute.
        Handles the fact, that the transitions are distributed in blocks.

        Parameters
        ----------
        _key : str
            attribute name for the A_mytx data array
        key : str
            attribute name for the A_tx data array
        """
        # In general, the data array has not been specified to instances of
        # the class. As a result, set the _key to None
        setattr(cls, _key, None)
        # self.key should return full data array
        setattr(cls, key,
                property(lambda self: self.get_all(self.__dict__[_key])))

    def attach(self, _key, key, A_mytx):
        """Attach a data array to the k-point pair.
        Used by PairMatrixElement to attach matrix elements calculated
        between the k-points for the different transitions."""
        self.add_mytransitions_array(_key, key)
        setattr(self, _key, A_mytx)


class KohnShamPair:
    """Class for extracting pairs of Kohn-Sham orbitals from a ground
    state calculation."""

    def __init__(self, gs, world=mpi.world, transitionblockscomm=None,
                 kptblockcomm=None, txt='-', timer=None):
        """
        Parameters
        ----------
        transitionblockscomm : gpaw.mpi.Communicator
            Communicator for distributing the transitions among processes
        kptblockcomm : gpaw.mpi.Communicator
            Communicator for distributing k-points among processes
        """
        self.world = world
        self.fd = convert_string_to_fd(txt, world)
        self.timer = timer or Timer()
        self.calc = get_calc(gs, fd=self.fd, timer=self.timer)
        self.calc_parallel = self.check_calc_parallelisation()

        self.transitionblockscomm = transitionblockscomm
        self.kptblockcomm = kptblockcomm

        # Prepare to distribute transitions
        self.mynt = None
        self.nt = None
        self.ta = None
        self.tb = None

        # Prepare to find k-point data from vector
        kd = self.calc.wfs.kd
        self.kdtree = cKDTree(np.mod(np.mod(kd.bzk_kc, 1).round(6), 1))

        # Prepare to use other processes' k-points
        self._pd0 = None

        # Prepare to redistribute kptdata
        self.rrequests = []
        self.srequests = []

        # Count bands so it is possible to remove null transitions
        self.nocc1 = None  # number of completely filled bands
        self.nocc2 = None  # number of non-empty bands
        self.count_occupied_bands()

    def check_calc_parallelisation(self):
        """Check how ground state calculation is distributed in memory"""
        if self.calc.world.size == 1:
            return False
        else:
            assert self.world.rank == self.calc.wfs.world.rank
            assert self.calc.wfs.gd.comm.size == 1
            return True

    def count_occupied_bands(self):
        """Count number of occupied and unoccupied bands in ground state
        calculation. Can be used to omit null-transitions between two occupied
        bands or between two unoccupied bands."""
        ftol = 1.e-9  # Could be given as input
        nocc1 = 9999999
        nocc2 = 0
        for kpt in self.calc.wfs.kpt_u:
            f_n = kpt.f_n / kpt.weight
            nocc1 = min((f_n > 1 - ftol).sum(), nocc1)
            nocc2 = max((f_n > ftol).sum(), nocc2)
        nocc1 = int(nocc1)
        nocc2 = int(nocc2)

        # Collect nocc for all k-points
        nocc1 = self.calc.wfs.kd.comm.min(nocc1)
        nocc2 = self.calc.wfs.kd.comm.max(nocc2)

        # Sum over band distribution
        nocc1 = self.calc.wfs.bd.comm.sum(nocc1)
        nocc2 = self.calc.wfs.bd.comm.sum(nocc2)

        self.nocc1 = int(nocc1)
        self.nocc2 = int(nocc2)
        print('Number of completely filled bands:', self.nocc1, file=self.fd)
        print('Number of partially filled bands:', self.nocc2, file=self.fd)
        print('Total number of bands:', self.calc.wfs.bd.nbands,
              file=self.fd)

    @property
    def pd0(self):
        """Get a PWDescriptor that includes all k-points"""
        if self._pd0 is None:
            from gpaw.wavefunctions.pw import PWDescriptor
            wfs = self.calc.wfs
            assert wfs.gd.comm.size == 1

            kd0 = wfs.kd.copy()
            pd, gd = wfs.pd, wfs.gd

            # Extract stuff from self.calc.wfs.pd
            ecut, dtype = pd.ecut, pd.dtype
            fftwflags, gammacentered = pd.fftwflags, pd.gammacentered

            # Initiate _pd0 with kd0
            self._pd0 = PWDescriptor(ecut, gd, dtype=dtype,
                                     kd=kd0, fftwflags=fftwflags,
                                     gammacentered=gammacentered)
        return self._pd0

    @timer('Get Kohn-Sham pairs')
    def get_kpoint_pairs(self, n1_t, n2_t, k1_pc, k2_pc, s1_t, s2_t):
        """Get all pairs of Kohn-Sham orbitals for transitions:
        (n1_t, k1_p, s1_t) -> (n2_t, k2_p, s2_t)
        Here, t is a composite band and spin transition index
        and p is indexing the different k-points to be distributed."""

        # Distribute transitions and extract data for transitions in
        # this process' block
        nt = len(n1_t)
        assert nt == len(n2_t)
        self.distribute_transitions(nt)

        kpt1 = self.get_kpoints(k1_pc, n1_t, s1_t)
        kpt2 = self.get_kpoints(k2_pc, n2_t, s2_t)

        # The process might not have any k-point pairs to evaluate, as
        # their are distributed in the kptblockcomm
        if kpt1 is None:
            assert kpt2 is None
            return None
        assert kpt2 is not None

        return KohnShamKPointPair(kpt1, kpt2,
                                  self.mynt, nt, self.ta, self.tb,
                                  comm=self.transitionblockscomm)

    def distribute_transitions(self, nt):
        """Distribute transitions between processes in block communicator"""
        if self.transitionblockscomm is None:
            mynt = nt
            ta = 0
            tb = nt
        else:
            nblocks = self.transitionblockscomm.size
            rank = self.transitionblockscomm.rank

            mynt = (nt + nblocks - 1) // nblocks
            ta = min(rank * mynt, nt)
            tb = min(ta + mynt, nt)

        self.mynt = mynt
        self.nt = nt
        self.ta = ta
        self.tb = tb

    def get_kpoints(self, k_pc, n_t, s_t):
        """Get KohnShamKPoint and help other processes extract theirs"""
        assert len(n_t) == len(s_t)
        assert len(k_pc) <= self.kptblockcomm.size
        kpt = None

        # Use the data extraction factory to extract the kptdata
        _extract_kptdata = self.create_extract_kptdata()
        kptdata = _extract_kptdata(k_pc, n_t, s_t)

        # Make local n and s arrays for the KohnShamKPoint object
        n_myt = np.empty(self.mynt, dtype=n_t.dtype)
        n_myt[:self.tb - self.ta] = n_t[self.ta:self.tb]
        s_myt = np.empty(self.mynt, dtype=s_t.dtype)
        s_myt[:self.tb - self.ta] = s_t[self.ta:self.tb]

        # Initiate k-point object.
        if self.kptblockcomm.rank in range(len(k_pc)):
            assert kptdata is not None
            kpt = KohnShamKPoint(n_myt, s_myt, *kptdata)

        return kpt

    def create_extract_kptdata(self):
        """Creator component of the data extraction factory."""
        if self.calc_parallel:
            return self.parallel_extract_kptdata
        else:
            return self.serial_extract_kptdata
            # Useful for debugging:
            # return self.parallel_extract_kptdata

    def parallel_extract_kptdata(self, k_pc, n_t, s_t):
        """Returns the input to KohnShamKPoint:
        K, n_myt, s_myt, eps_myt, f_myt, ut_mytR, projections, shift_c
        if a k-point in the given list, k_pc, belongs to the process.
        """
        # Extract the data from the ground state calculator object
        data, h_myt, myt_myt = self._parallel_extract_kptdata(k_pc, n_t, s_t)

        # If the process has a k-point to return, symmetrize and unfold
        if self.kptblockcomm.rank in range(len(k_pc)):
            assert data is not None
            # Unpack data, apply FT and symmetrization
            K, k_c, eps_h, f_h, Ph, psit_hG = data
            Ph, ut_hR, shift_c = self.transform_and_symmetrize(K, k_c, Ph,
                                                               psit_hG)

            (eps_myt, f_myt,
             P, ut_mytR) = self.unfold_arrays(eps_h, f_h, Ph, ut_hR,
                                              h_myt, myt_myt)

            data = (K, eps_myt, f_myt, ut_mytR, P, shift_c)

        # Wait for communication to finish
        with self.timer('Waiting to complete mpi.send'):
            while self.srequests:
                self.world.wait(self.srequests.pop(0))

        return data

    @timer('Extracting data from the ground state calculator object')
    def _parallel_extract_kptdata(self, k_pc, n_t, s_t):
        """In-place kptdata extraction."""
        (data, myu_eu,
         myn_eueh, ik_r2,
         nrh_r2, eh_eur2reh,
         rh_eur2reh, h_r1rh,
         h_myt, myt_myt) = self.get_extraction_protocol(k_pc, n_t, s_t)

        (eps_r1rh, f_r1rh,
         P_r1rhI, psit_r1rhG,
         eps_r2rh, f_r2rh,
         P_r2rhI, psit_r2rhG) = self.allocate_transfer_arrays(data, nrh_r2,
                                                              ik_r2, h_r1rh)

        # Do actual extraction
        for myu, myn_eh, eh_r2reh, rh_r2reh in zip(myu_eu, myn_eueh,
                                                   eh_eur2reh, rh_eur2reh):

            eps_eh, f_eh, P_ehI = self.extract_wfs_data(myu, myn_eh)

            for r2, (eh_reh, rh_reh) in enumerate(zip(eh_r2reh, rh_r2reh)):
                if eh_reh:
                    eps_r2rh[r2][rh_reh] = eps_eh[eh_reh]
                    f_r2rh[r2][rh_reh] = f_eh[eh_reh]
                    P_r2rhI[r2][rh_reh] = P_ehI[eh_reh]

            # Wavefunctions are heavy objects which can only be extracted
            # for one band index at a time, handle them seperately
            self.add_wave_function(myu, myn_eh, eh_r2reh,
                                   rh_r2reh, psit_r2rhG)

        self.distribute_extracted_data(eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG,
                                       eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG)

        data = self.collect_kptdata(data, h_r1rh, eps_r1rh,
                                    f_r1rh, P_r1rhI, psit_r1rhG)

        return data, h_myt, myt_myt

    @timer('Create data extraction protocol')
    def get_extraction_protocol(self, k_pc, n_t, s_t):
        """Figure out how to extract data efficiently.
        For the serial communicator, all processes can access all data,
        and resultantly, there is no need to send any data.
        """
        wfs = self.calc.wfs
        get_extraction_info = self.create_get_extraction_info()

        # Kpoint data
        data = (None, None, None)

        # Extraction protocol
        myu_eu = []
        myn_eueh = []

        # Data distribution protocol
        nrh_r2 = np.zeros(self.world.size, dtype=int)
        ik_r2 = [None for _ in range(self.world.size)]
        eh_eur2reh = []
        rh_eur2reh = []
        h_r1rh = [list([]) for _ in range(self.world.size)]

        # h to t index mapping
        myt_myt = np.arange(self.tb - self.ta)
        t_myt = range(self.ta, self.tb)
        n_myt, s_myt = n_t[t_myt], s_t[t_myt]
        h_myt = np.empty(self.tb - self.ta, dtype=int)

        nt = len(n_t)
        assert nt == len(s_t)
        t_t = np.arange(nt)
        nh = 0
        for p, k_c in enumerate(k_pc):  # p indicates the receiving process
            K = self.find_kpoint(k_c)
            ik = wfs.kd.bz2ibz_k[K]
            for r2 in range(p * self.transitionblockscomm.size,
                            min((p + 1) * self.transitionblockscomm.size,
                                self.world.size)):
                ik_r2[r2] = ik

            if p == self.kptblockcomm.rank:
                data = (K, k_c, ik)

            # Find out who should store the data in KSKPpoint
            r2_t, myt_t = self.map_who_has(p, t_t)

            # Find out how to extract data
            # In the ground state, kpts are indexed by u=(s, k)
            for s in set(s_t):
                thiss_myt = s_myt == s
                thiss_t = s_t == s
                t_ct = t_t[thiss_t]
                n_ct = n_t[thiss_t]
                r2_ct = r2_t[t_ct]

                # Find out where data is in wfs
                u = ik * wfs.nspins + s
                myu, r1_ct, myn_ct = get_extraction_info(u, n_ct, r2_ct)

                # If the process is extracting or receiving data,
                # figure out how to do so
                if self.world.rank in np.append(r1_ct, r2_ct):
                    # Does this process have anything to send?
                    thisr1_ct = r1_ct == self.world.rank
                    if np.any(thisr1_ct):
                        eh_r2reh = [list([]) for _ in range(self.world.size)]
                        rh_r2reh = [list([]) for _ in range(self.world.size)]
                        # Find composite indeces h = (n, s)
                        n_et = n_ct[thisr1_ct]
                        n_eh = np.unique(n_et)
                        # Find composite local band indeces
                        myn_eh = np.unique(myn_ct[thisr1_ct])

                        # Where to send the data
                        r2_et = r2_ct[thisr1_ct]
                        for r2 in np.unique(r2_et):
                            thisr2_et = r2_et == r2
                            # What ns are the process sending?
                            n_reh = np.unique(n_et[thisr2_et])
                            eh_reh = []
                            for n in n_reh:
                                eh_reh.append(np.where(n_eh == n)[0][0])
                            # How to send it
                            eh_r2reh[r2] = eh_reh
                            nreh = len(eh_reh)
                            rh_r2reh[r2] = np.arange(nreh) + nrh_r2[r2]
                            nrh_r2[r2] += nreh

                        myu_eu.append(myu)
                        myn_eueh.append(myn_eh)
                        eh_eur2reh.append(eh_r2reh)
                        rh_eur2reh.append(rh_r2reh)

                    # Does this process have anything to receive?
                    thisr2_ct = r2_ct == self.world.rank
                    if np.any(thisr2_ct):
                        # Find unique composite indeces h = (n, s)
                        n_rt = n_ct[thisr2_ct]
                        n_rn = np.unique(n_rt)
                        nrn = len(n_rn)
                        h_rn = np.arange(nrn) + nh
                        nh += nrn

                        # Where to get the data from
                        r1_rt = r1_ct[thisr2_ct]
                        for r1 in np.unique(r1_rt):
                            thisr1_rt = r1_rt == r1
                            # What ns are the process getting?
                            n_reh = np.unique(n_rt[thisr1_rt])
                            # Where to put them
                            for n in n_reh:
                                h = h_rn[np.where(n_rn == n)[0][0]]
                                h_r1rh[r1].append(h)

                                # h to t mapping
                                thisn_myt = n_myt == n
                                thish_myt = np.logical_and(thisn_myt,
                                                           thiss_myt)
                                h_myt[thish_myt] = h

        return (data, myu_eu, myn_eueh, ik_r2, nrh_r2,
                eh_eur2reh, rh_eur2reh, h_r1rh, h_myt, myt_myt)

    def create_get_extraction_info(self):
        """Creator component of the extraction information factory."""
        if self.calc_parallel:
            return self.get_parallel_extraction_info
        else:
            return self.get_serial_extraction_info

    @staticmethod
    def get_serial_extraction_info(u, n_ct, r2_ct):
        """Figure out where to extract the data from in the gs calc"""
        # Let the process extract its own data
        myu = u  # The process has access to all data
        r1_ct = r2_ct
        myn_ct = n_ct

        return myu, r1_ct, myn_ct

    def get_parallel_extraction_info(self, u, n_ct, *unused):
        """Figure out where to extract the data from in the gs calc"""
        wfs = self.calc.wfs
        # Find out where data is in wfs
        k, s = divmod(u, wfs.nspins)
        kptrank, q = wfs.kd.who_has(k)
        myu = q * wfs.nspins + s
        r1_ct, myn_ct = [], []
        for n in n_ct:
            bandrank, myn = wfs.bd.who_has(n)
            # XXX this will fail when using non-standard nesting
            # of communicators.
            r1 = (kptrank * wfs.gd.comm.size * wfs.bd.comm.size
                  + bandrank * wfs.gd.comm.size)
            r1_ct.append(r1)
            myn_ct.append(myn)

        return myu, np.array(r1_ct), np.array(myn_ct)

    @timer('Allocate transfer arrays')
    def allocate_transfer_arrays(self, data, nrh_r2, ik_r2, h_r1rh):
        """Allocate arrays for intermediate storage of data."""
        wfs = self.calc.wfs
        kptex = wfs.kpt_u[0]
        Pshape = kptex.projections.array.shape
        Pdtype = kptex.projections.matrix.dtype
        psitdtype = kptex.psit.array.dtype

        # Number of h-indeces to receive
        nrh_r1 = [len(h_rh) for h_rh in h_r1rh]

        # if self.kptblockcomm.rank in range(len(ik_p)):
        if data[2] is not None:
            ik = data[2]
            ng = self.pd0.ng_q[ik]
            eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG = [], [], [], []
            for nrh in nrh_r1:
                if nrh >= 1:
                    eps_r1rh.append(np.empty(nrh))
                    f_r1rh.append(np.empty(nrh))
                    P_r1rhI.append(np.empty((nrh,) + Pshape[1:], dtype=Pdtype))
                    psit_r1rhG.append(np.empty((nrh, ng), dtype=psitdtype))
                else:
                    eps_r1rh.append(None)
                    f_r1rh.append(None)
                    P_r1rhI.append(None)
                    psit_r1rhG.append(None)
        else:
            eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG = None, None, None, None

        eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG = [], [], [], []
        for nrh, ik in zip(nrh_r2, ik_r2):
            if nrh:
                eps_r2rh.append(np.empty(nrh))
                f_r2rh.append(np.empty(nrh))
                P_r2rhI.append(np.empty((nrh,) + Pshape[1:], dtype=Pdtype))
                ng = self.pd0.ng_q[ik]
                psit_r2rhG.append(np.empty((nrh, ng), dtype=psitdtype))
            else:
                eps_r2rh.append(None)
                f_r2rh.append(None)
                P_r2rhI.append(None)
                psit_r2rhG.append(None)

        return (eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG,
                eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG)

    def map_who_has(self, p, t_t):
        """Convert k-point and transition index to global world rank
        and local transition index"""
        trank_t, myt_t = np.divmod(t_t, self.mynt)
        return p * self.transitionblockscomm.size + trank_t, myt_t

    @timer('Extracting eps, f and P_I from wfs')
    def extract_wfs_data(self, myu, myn_eh):
        wfs = self.calc.wfs
        kpt = wfs.kpt_u[myu]
        # Get eig and occ
        eps_eh, f_eh = kpt.eps_n[myn_eh], kpt.f_n[myn_eh] / kpt.weight

        # Get projections
        assert kpt.projections.atom_partition.comm.size == 1
        P_ehI = kpt.projections.array[myn_eh]

        return eps_eh, f_eh, P_ehI

    @timer('Extracting wave function from wfs')
    def add_wave_function(self, myu, myn_eh,
                          eh_r2reh, rh_r2reh, psit_r2rhG):
        """Add the plane wave coefficients of the smooth part of
        the wave function to the psit_r2rtG arrays."""
        wfs = self.calc.wfs
        kpt = wfs.kpt_u[myu]

        for eh_reh, rh_reh, psit_rhG in zip(eh_r2reh, rh_r2reh, psit_r2rhG):
            if eh_reh:
                for eh, rh in zip(eh_reh, rh_reh):
                    psit_rhG[rh] = kpt.psit_nG[myn_eh[eh]]

    @timer('Distributing kptdata')
    def distribute_extracted_data(self, eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG,
                                  eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG):
        """Send the extracted data to appropriate destinations"""
        # Store the data extracted by the process itself
        rank = self.world.rank
        # Check if there is actually some data to store
        if eps_r2rh[rank] is not None:
            eps_r1rh[rank] = eps_r2rh[rank]
            f_r1rh[rank] = f_r2rh[rank]
            P_r1rhI[rank] = P_r2rhI[rank]
            psit_r1rhG[rank] = psit_r2rhG[rank]

        # Receive data
        if eps_r1rh is not None:  # The process may not be receiving anything
            for r1, (eps_rh, f_rh,
                     P_rhI, psit_rhG) in enumerate(zip(eps_r1rh, f_r1rh,
                                                       P_r1rhI, psit_r1rhG)):
                # Check if there is any data to receive
                if r1 != rank and eps_rh is not None:
                    rreq1 = self.world.receive(eps_rh, r1,
                                               tag=201, block=False)
                    rreq2 = self.world.receive(f_rh, r1,
                                               tag=202, block=False)
                    rreq3 = self.world.receive(P_rhI, r1,
                                               tag=203, block=False)
                    rreq4 = self.world.receive(psit_rhG, r1,
                                               tag=204, block=False)
                    self.rrequests += [rreq1, rreq2, rreq3, rreq4]

        # Send data
        for r2, (eps_rh, f_rh,
                 P_rhI, psit_rhG) in enumerate(zip(eps_r2rh, f_r2rh,
                                                   P_r2rhI, psit_r2rhG)):
            # Check if there is any data to send
            if r2 != rank and eps_rh is not None:
                sreq1 = self.world.send(eps_rh, r2, tag=201, block=False)
                sreq2 = self.world.send(f_rh, r2, tag=202, block=False)
                sreq3 = self.world.send(P_rhI, r2, tag=203, block=False)
                sreq4 = self.world.send(psit_rhG, r2, tag=204, block=False)
                self.srequests += [sreq1, sreq2, sreq3, sreq4]

        with self.timer('Waiting to complete mpi.receive'):
            while self.rrequests:
                self.world.wait(self.rrequests.pop(0))

    @timer('Collecting kptdata')
    def collect_kptdata(self, data, h_r1rh,
                        eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG):
        """From the extracted data, collect the KohnShamKPoint data arrays"""

        # Some processes may not have to return a k-point
        if data[0] is None:
            return None
        K, k_c, ik = data

        # Allocate data arrays
        wfs = self.calc.wfs
        maxh_r1 = [max(h_rh) for h_rh in h_r1rh if h_rh]
        if maxh_r1:
            nh = max(maxh_r1) + 1
        else:  # Carry around empty array
            assert self.ta == self.tb
            nh = 1
        eps_h = np.empty(nh)
        f_h = np.empty(nh)
        Ph = wfs.kpt_u[0].projections.new(nbands=nh, bcomm=None)
        psit_hG = np.empty((nh, self.pd0.ng_q[ik]),
                           dtype=wfs.kpt_u[0].psit.array.dtype)

        # Store extracted data in the arrays
        for (h_rh, eps_rh,
             f_rh, P_rhI, psit_rhG) in zip(h_r1rh, eps_r1rh,
                                           f_r1rh, P_r1rhI, psit_r1rhG):
            if h_rh:
                eps_h[h_rh] = eps_rh
                f_h[h_rh] = f_rh
                Ph.array[h_rh] = P_rhI
                psit_hG[h_rh] = psit_rhG

        return (K, k_c, eps_h, f_h, Ph, psit_hG)

    @timer('Unfolding arrays')
    def unfold_arrays(self, eps_h, f_h, Ph, ut_hR, h_myt, myt_myt):
        """Create transition data arrays from the composite h = (n, s) index"""

        wfs = self.calc.wfs
        # Allocate data arrays for the k-point
        mynt = self.mynt
        eps_myt = np.empty(mynt)
        f_myt = np.empty(mynt)
        P = wfs.kpt_u[0].projections.new(nbands=mynt, bcomm=None)
        ut_mytR = wfs.gd.empty(self.mynt, wfs.dtype)

        # Unfold k-point data
        eps_myt[myt_myt] = eps_h[h_myt]
        f_myt[myt_myt] = f_h[h_myt]
        P.array[myt_myt] = Ph.array[h_myt]
        ut_mytR[myt_myt] = ut_hR[h_myt]

        return eps_myt, f_myt, P, ut_mytR

    @timer('Extracting data from the ground state calculator object')
    def serial_extract_kptdata(self, k_pc, n_t, s_t):
        # All processes can access all data. Each process extracts it own data.
        wfs = self.calc.wfs

        # Do data extraction for the processes, which have data to extract
        if self.kptblockcomm.rank in range(len(k_pc)):
            # Find k-point indeces
            k_c = k_pc[self.kptblockcomm.rank]
            K = self.find_kpoint(k_c)
            ik = wfs.kd.bz2ibz_k[K]
            # Construct symmetry operators
            (_, T, a_a, U_aii, shift_c,
             time_reversal) = self.construct_symmetry_operators(K, k_c=k_c)

            (myu_eu, myn_eurn, nh, h_eurn, h_myt,
             myt_myt) = self.get_serial_extraction_protocol(ik, n_t, s_t)

            # Allocate transfer arrays
            eps_h = np.empty(nh)
            f_h = np.empty(nh)
            Ph = wfs.kpt_u[0].projections.new(nbands=nh, bcomm=None)
            ut_hR = wfs.gd.empty(nh, wfs.dtype)

            # Extract data from the ground state
            for myu, myn_rn, h_rn in zip(myu_eu, myn_eurn, h_eurn):
                kpt = wfs.kpt_u[myu]
                with self.timer('Extracting eps, f and P_I from wfs'):
                    eps_h[h_rn] = kpt.eps_n[myn_rn]
                    f_h[h_rn] = kpt.f_n[myn_rn] / kpt.weight
                    Ph.array[h_rn] = kpt.projections.array[myn_rn]

                with self.timer('Extracting, fourier transforming and '
                                'symmetrizing wave function'):
                    for myn, h in zip(myn_rn, h_rn):
                        ut_hR[h] = T(wfs.pd.ifft(kpt.psit_nG[myn], kpt.q))

            # Symmetrize projections
            with self.timer('Apply symmetry operations'):
                P_ahi = []
                for a1, U_ii in zip(a_a, U_aii):
                    P_hi = np.ascontiguousarray(Ph[a1])
                    # Apply symmetry operations. This will map a1 onto a2
                    np.dot(P_hi, U_ii, out=P_hi)
                    if time_reversal:
                        np.conj(P_hi, out=P_hi)
                    P_ahi.append(P_hi)

                # Store symmetrized projectors
                for a2, P_hi in enumerate(P_ahi):
                    I1, I2 = Ph.map[a2]
                    Ph.array[..., I1:I2] = P_hi

            (eps_myt, f_myt,
             P, ut_mytR) = self.unfold_arrays(eps_h, f_h, Ph, ut_hR,
                                              h_myt, myt_myt)

            return (K, eps_myt, f_myt, ut_mytR, P, shift_c)

    @timer('Create data extraction protocol')
    def get_serial_extraction_protocol(self, ik, n_t, s_t):
        """Figure out how to extract data efficiently.
        For the serial communicator, all processes can access all data,
        and resultantly, there is no need to send any data.
        """
        wfs = self.calc.wfs

        # Only extract the transitions handled by the process itself
        myt_myt = np.arange(self.tb - self.ta)
        t_myt = range(self.ta, self.tb)
        n_myt = n_t[t_myt]
        s_myt = s_t[t_myt]

        # In the ground state, kpts are indexed by u=(s, k)
        myu_eu = []
        myn_eurn = []
        nh = 0
        h_eurn = []
        h_myt = np.empty(self.tb - self.ta, dtype=int)
        for s in set(s_myt):
            thiss_myt = s_myt == s
            n_ct = n_myt[thiss_myt]

            # Find unique composite h = (n, u) indeces
            n_rn = np.unique(n_ct)
            nrn = len(n_rn)
            h_eurn.append(np.arange(nrn) + nh)
            nh += nrn

            # Find mapping between h and the transition index
            for n, h in zip(n_rn, h_eurn[-1]):
                thisn_myt = n_myt == n
                thish_myt = np.logical_and(thisn_myt, thiss_myt)
                h_myt[thish_myt] = h

            # Find out where data is in wfs
            u = ik * wfs.nspins + s
            # The process has access to all data
            myu = u
            myn_rn = n_rn

            myu_eu.append(myu)
            myn_eurn.append(myn_rn)

        return myu_eu, myn_eurn, nh, h_eurn, h_myt, myt_myt

    @timer('Identifying k-points')
    def find_kpoint(self, k_c):
        return self.kdtree.query(np.mod(np.mod(k_c, 1).round(6), 1))[1]

    @timer('Apply symmetry operations')
    def transform_and_symmetrize(self, K, k_c, Ph, psit_hG):
        """Get wave function on a real space grid and symmetrize it
        along with the corresponding PAW projections."""
        (_, T, a_a, U_aii, shift_c,
         time_reversal) = self.construct_symmetry_operators(K, k_c=k_c)

        # Symmetrize wave functions
        wfs = self.calc.wfs
        ik = wfs.kd.bz2ibz_k[K]
        ut_hR = wfs.gd.empty(len(psit_hG), wfs.dtype)
        with self.timer('Fourier transform and symmetrize wave functions'):
            for h, psit_G in enumerate(psit_hG):
                ut_hR[h] = T(self.pd0.ifft(psit_G, ik))

        # Symmetrize projections
        P_ahi = []
        for a1, U_ii in zip(a_a, U_aii):
            P_hi = np.ascontiguousarray(Ph[a1])
            # Apply symmetry operations. This will map a1 onto a2
            np.dot(P_hi, U_ii, out=P_hi)
            if time_reversal:
                np.conj(P_hi, out=P_hi)
            P_ahi.append(P_hi)

        # Store symmetrized projectors
        for a2, P_hi in enumerate(P_ahi):
            I1, I2 = Ph.map[a2]
            Ph.array[..., I1:I2] = P_hi

        return Ph, ut_hR, shift_c

    @timer('Construct symmetry operators')
    def construct_symmetry_operators(self, K, k_c=None):
        """Construct symmetry operators for wave function and PAW projections.

        We want to transform a k-point in the irreducible part of the BZ to
        the corresponding k-point with index K.

        Returns U_cc, T, a_a, U_aii, shift_c and time_reversal, where:

        * U_cc is a rotation matrix.
        * T() is a function that transforms the periodic part of the wave
          function.
        * a_a is a list of symmetry related atom indices
        * U_aii is a list of rotation matrices for the PAW projections
        * shift_c is three integers: see code below.
        * time_reversal is a flag - if True, projections should be complex
          conjugated.

        See the extract_orbitals() method for how to use these tuples.
        """

        wfs = self.calc.wfs
        kd = wfs.kd

        s = kd.sym_k[K]
        U_cc = kd.symmetry.op_scc[s]
        time_reversal = kd.time_reversal_k[K]
        ik = kd.bz2ibz_k[K]
        if k_c is None:
            k_c = kd.bzk_kc[K]
        ik_c = kd.ibzk_kc[ik]

        sign = 1 - 2 * time_reversal
        shift_c = np.dot(U_cc, ik_c) - k_c * sign

        try:
            assert np.allclose(shift_c.round(), shift_c)
        except AssertionError:
            print('shift_c ' + str(shift_c), file=self.fd)
            print('k_c ' + str(k_c), file=self.fd)
            print('kd.bzk_kc[K] ' + str(kd.bzk_kc[K]), file=self.fd)
            print('ik_c ' + str(ik_c), file=self.fd)
            print('U_cc ' + str(U_cc), file=self.fd)
            print('sign ' + str(sign), file=self.fd)
            raise AssertionError

        shift_c = shift_c.round().astype(int)

        if (U_cc == np.eye(3)).all():
            def T(f_R):
                return f_R
        else:
            N_c = self.calc.wfs.gd.N_c
            i_cr = np.dot(U_cc.T, np.indices(N_c).reshape((3, -1)))
            i = np.ravel_multi_index(i_cr, N_c, 'wrap')

            def T(f_R):
                return f_R.ravel()[i].reshape(N_c)

        if time_reversal:
            T0 = T

            def T(f_R):
                return T0(f_R).conj()

            shift_c *= -1

        a_a = []
        U_aii = []
        for a, id in enumerate(self.calc.wfs.setups.id_a):
            b = kd.symmetry.a_sa[s, a]
            S_c = np.dot(self.calc.spos_ac[a], U_cc) - self.calc.spos_ac[b]
            x = np.exp(2j * np.pi * np.dot(ik_c, S_c))
            U_ii = wfs.setups[a].R_sii[s].T * x
            a_a.append(b)
            U_aii.append(U_ii)

        shift0_c = (kd.bzk_kc[K] - k_c).round().astype(int)
        shift_c += -shift0_c

        return U_cc, T, a_a, U_aii, shift_c, time_reversal


def get_calc(gs, fd=None, timer=None):
    """Get ground state calculation object."""
    if isinstance(gs, GPAW):
        return gs
    else:
        if timer is None:
            def timer(*unused):
                def __enter__(self):
                    pass

                def __exit__(self):
                    pass

        with timer('Read ground state'):
            assert Path(gs).is_file()
            if fd is not None:
                print('Reading ground state calculation:\n  %s' % gs,
                      file=fd)
            return GPAW(gs, txt=None, communicator=mpi.serial_comm)


class PairMatrixElement:
    """Class for calculating matrix elements for transitions in Kohn-Sham
    linear response functions."""
    def __init__(self, kspair):
        """
        Parameters
        ----------
        kslrf : KohnShamLinearResponseFunction instance
        """
        self.calc = kspair.calc
        self.fd = kspair.fd
        self.timer = kspair.timer
        self.transitionblockscomm = kspair.transitionblockscomm

    def initialize(self, *args, **kwargs):
        """Initialize e.g. PAW corrections or other operations
        ahead in time of integration."""
        pass

    def __call__(self, kskptpair, *args, **kwargs):
        """Calculate the matrix element for all transitions in kskptpairs."""
        raise NotImplementedError('Define specific matrix element')


class PlaneWavePairDensity(PairMatrixElement):
    """Class for calculating pair densities:

    n_T(q+G) = <s'n'k'| e^(i (q + G) r) |snk>

    in the plane wave mode"""
    def __init__(self, kspair):
        PairMatrixElement.__init__(self, kspair)

        # Save PAW correction for all calls with same q_c
        self.Q_aGii = None
        self.currentq_c = None

    def initialize(self, pd):
        """Initialize PAW corrections ahead in time of integration."""
        self.initialize_paw_corrections(pd)

    @timer('Initialize PAW corrections')
    def initialize_paw_corrections(self, pd):
        """Initialize PAW corrections, if not done already, for the given q"""
        q_c = pd.kd.bzk_kc[0]
        if self.Q_aGii is None or not np.allclose(q_c - self.currentq_c, 0.):
            self.Q_aGii = self._initialize_paw_corrections(pd)
            self.currentq_c = q_c

    def _initialize_paw_corrections(self, pd):
        wfs = self.calc.wfs
        spos_ac = self.calc.spos_ac
        G_Gv = pd.get_reciprocal_vectors()

        pos_av = np.dot(spos_ac, pd.gd.cell_cv)

        # Collect integrals for all species:
        Q_xGii = {}
        for id, atomdata in wfs.setups.setups.items():
            Q_Gii = two_phi_planewave_integrals(G_Gv, atomdata)
            ni = atomdata.ni
            Q_Gii.shape = (-1, ni, ni)

            Q_xGii[id] = Q_Gii

        Q_aGii = []
        for a, atomdata in enumerate(wfs.setups):
            id = wfs.setups.id_a[a]
            Q_Gii = Q_xGii[id]
            x_G = np.exp(-1j * np.dot(G_Gv, pos_av[a]))
            Q_aGii.append(x_G[:, np.newaxis, np.newaxis] * Q_Gii)

        return Q_aGii

    @timer('Calculate pair density')
    def __call__(self, kskptpair, pd):
        """Calculate the pair densities for all transitions:
        n_t(q+G) = <s'n'k+q| e^(i (q + G) r) |snk>
                 = <snk| e^(-i (q + G) r) |s'n'k+q>
        """
        Q_aGii = self.get_paw_projectors(pd)
        Q_G = self.get_fft_indices(kskptpair, pd)
        mynt, nt, ta, tb = kskptpair.transition_distribution()

        n_mytG = pd.empty(mynt)

        # Calculate smooth part of the pair densities:
        with self.timer('Calculate smooth part'):
            ut1cc_mytR = kskptpair.kpt1.ut_tR.conj()
            n_mytR = ut1cc_mytR * kskptpair.kpt2.ut_tR
            # Unvectorized
            for myt in range(tb - ta):
                n_mytG[myt] = pd.fft(n_mytR[myt], 0, Q_G) * pd.gd.dv

        # Calculate PAW corrections with numpy
        with self.timer('PAW corrections'):
            P1 = kskptpair.kpt1.projections
            P2 = kskptpair.kpt2.projections
            for (Q_Gii, (a1, P1_myti),
                 (a2, P2_myti)) in zip(Q_aGii, P1.items(), P2.items()):
                P1cc_myti = P1_myti[:tb - ta].conj()
                C1_Gimyt = np.tensordot(Q_Gii, P1cc_myti, axes=([1, 1]))
                P2_imyt = P2_myti.T[:, :tb - ta]
                n_mytG[:tb - ta] += np.sum(C1_Gimyt * P2_imyt[np.newaxis,
                                                              :, :], axis=1).T

        # Attach the calculated pair density to the KohnShamKPointPair object
        kskptpair.attach('n_mytG', 'n_tG', n_mytG)

    def get_paw_projectors(self, pd):
        """Make sure PAW correction has been initialized properly
        and return projectors"""
        self.initialize_paw_corrections(pd)
        return self.Q_aGii

    @timer('Get G-vector indices')
    def get_fft_indices(self, kskptpair, pd):
        """Get indices for G-vectors inside cutoff sphere."""
        kpt1 = kskptpair.kpt1
        kpt2 = kskptpair.kpt2
        kd = self.calc.wfs.kd
        q_c = pd.kd.bzk_kc[0]

        N_G = pd.Q_qG[0]

        shift_c = kpt1.shift_c - kpt2.shift_c
        shift_c += (q_c - kd.bzk_kc[kpt2.K]
                    + kd.bzk_kc[kpt1.K]).round().astype(int)
        if shift_c.any():
            n_cG = np.unravel_index(N_G, pd.gd.N_c)
            n_cG = [n_G + shift for n_G, shift in zip(n_cG, shift_c)]
            N_G = np.ravel_multi_index(n_cG, pd.gd.N_c, 'wrap')
        return N_G



from ase import Atoms
from gpaw import GPAW

d = 1.1   # bondlength of hydrogen molecule
a = 5.0   # sidelength of unit cell
c = a / 2
atoms = Atoms('CO',
              positions=[(c - d / 2, c, c),
                         (c + d / 2, c, c)],
              cell=(a, a, a))

calc = GPAW(nbands=5, h=0.2, txt=None)
atoms.calc = calc

# Start a calculation:
energy = atoms.get_potential_energy()

# Save wave functions:
calc.write('CO.gpw', mode='all')



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.wavefunctions.pw import PW

bulk_si = bulk('Si', a=5.42935602)
calc = GPAW(mode=PW(400.0),
            xc='LDA',
            occupations=FermiDirac(width=0.01),
            kpts={'size': (4, 4, 4), 'gamma': True},
            parallel={'domain': 1},
            txt='si.gs.txt')

bulk_si.calc = calc
E_lda = bulk_si.get_potential_energy()
calc.diagonalize_full_hamiltonian()
calc.write('si.lda_wfcs.gpw', mode='all')



from ase import Atoms
from ase.build import bulk
from gpaw import GPAW, FermiDirac, PW

cell = bulk('Ga', 'fcc', a=5.68).cell
a = Atoms('GaAs', cell=cell, pbc=True,
          scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))

calc = GPAW(mode=PW(600),
            xc='LDA',
            occupations=FermiDirac(width=0.01),
            convergence={'density': 1.e-6},
            symmetry='off',
            kpts={'size': (2, 2, 2), 'gamma': True},
            txt='gs_GaAs.txt')

a.calc = calc
a.get_potential_energy()
calc.write('GaAs.gpw', mode='all')



from ase.io import write
from gpaw import GPAW
calc = GPAW('anti.gpw')
atoms = calc.get_atoms()
up = calc.get_pseudo_density(0)
down = calc.get_pseudo_density(1)
zeta = (up - down) / (up + down)
write('magnetization.cube', atoms, data=zeta)



# necessary MKL libs
libraries += ['mkl_intel_lp64', 'mkl_sequential', 'mkl_core', 'mkl_avx2', 'mkl_def', 'svml']
# Now add a EasyBuild "cover-all-bases" library_dirs
library_dirs = os.getenv('LD_LIBRARY_PATH').split(':')
include_dirs = os.getenv('INCLUDE').split(':')



from types import SimpleNamespace

import numpy as np
import pytest
from ase.units import Bohr as bohr

from gpaw import GPAW
from gpaw.utilities.ps2ae import PS2AE
from gpaw.zero_field_splitting import (WaveFunctions,
                                       create_compensation_charge, zfs1)


@pytest.mark.serial
def test_zfs_o2(gpw_files):
    calc = GPAW(gpw_files['o2_pw_wfs'])
    wf1 = WaveFunctions.from_calc(calc, 0)
    wf1 = wf1.view(5, 7)
    cc = create_compensation_charge(wf1, calc.spos_ac)
    D1 = zfs1(wf1, wf1, cc)
    print(D1)

    converter = PS2AE(calc)
    psit_nR = np.array([converter.get_wave_function(n, ae=True) * bohr**1.5
                        for n in [5, 6]])
    wf2 = WaveFunctions(psit_nR, {}, 0, calc.setups, converter.gd)
    cc = SimpleNamespace(add=lambda a, b: None)
    D2 = zfs1(wf2, wf2, cc)
    print(D2)



# Creates: optimization.txt

from gpaw import restart
from ase.parallel import paropen as open
from ase.optimize import QuasiNewton


molecule, calc = restart('H2.gpw', txt='H2-relaxed.txt')

e2 = molecule.get_potential_energy()
d0 = molecule.get_distance(0, 1)

fd = open('optimization.txt', 'w')
print('experimental bond length:', file=fd)
print(f'hydrogen molecule energy: {e2:5.2f} eV', file=fd)
print(f'bondlength              : {d0:5.2f} Ang', file=fd)

# Find the theoretical bond length:
relax = QuasiNewton(molecule, logfile='qn.log')
relax.run(fmax=0.05)

e2 = molecule.get_potential_energy()
d0 = molecule.get_distance(0, 1)

print(file=fd)
print('PBE energy minimum:', file=fd)
print(f'hydrogen molecule energy: {e2:5.2f} eV', file=fd)
print(f'bondlength              : {d0:5.2f} Ang', file=fd)
fd.close()



# Creates: PES_fig.png

import numpy as np
import matplotlib.pyplot as plt

output = True

# H2O

filename = 'H2O'

data = np.loadtxt(filename + '-td' + '.dat')
be_t = data[:, 0]
f_t = data[:, 1]

data = np.loadtxt(filename + '-dos' + '.dat')
be_d = data[:, 0]
f_d = data[:, 1]

if output:
    fig = plt.figure(1, figsize=(14, 8))
    fig.subplots_adjust(left=0.08, right=0.98)

    plt.subplot(3, 1, 1)
    plt.bar(be_d - 0.1,
            f_d / f_d.max(),
            color='k',
            width=0.2,
            label='DFT eigenvalue')
    plt.bar(be_t - 0.1,
            f_t / f_t.max(),
            color='r',
            width=0.2,
            label='lr-TDDFT method')

    data = np.loadtxt(filename + '.dat')
    be_e = data[:, 0]
    f_e = 2.1 * data[:, 1]

    plt.plot(be_e, f_e, '-b', label='Exp')
    plt.text(6.5, 0.8, 'H$_2$O', fontsize=22)
    # xlabel('Binding energy [eV]', fontsize=20)
    plt.ylabel('Spectroscopic factor', fontsize=20)
    plt.axis((5, 45, 0, 2))
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=18)

etol = 1.0
peaks = np.array([12.8, 14.9, 18.5])
ube_t = np.unique(np.round(be_t, 1))[:len(peaks)]
ube_d = np.unique(np.round(be_d, 1))[:len(peaks)]

for t, d, ref in zip(ube_t, ube_d, peaks):
    assert abs(t - ref) < etol
    assert abs(t - d) < etol

# CO

filename = 'CO'

data = np.loadtxt(filename + '-td' + '.dat')
be_t = data[:, 0]
f_t = data[:, 1]

data = np.loadtxt(filename + '-dos' + '.dat')
be_d = data[:, 0]
f_d = data[:, 1]

if output:
    plt.subplot(3, 1, 2)
    plt.bar(be_d - 0.1,
            f_d / f_d.max(),
            color='k',
            width=0.2,
            label='DFT eigenvalue')
    plt.bar(be_t - 0.1,
            f_t / f_t.max(),
            color='r',
            width=0.2,
            label='lr-TDDFT method')

    data = np.loadtxt(filename + '.dat')
    be_e = data[:, 0]
    f_e = 2.1 * data[:, 1]

    plt.plot(be_e, f_e, '-b', label='Exp')
    plt.text(6.5, 0.8, 'CO', fontsize=22)
    # xlabel('Binding energy [eV]', fontsize=20)
    # ylabel('Spectroscopic factor', fontsize=20)
    plt.axis((5, 45, 0, 2))
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=18)

etol = 1.0
peaks = np.array([13.7, 17.7, 19.0])
ube_t = np.unique(np.round(be_t, 1))[:len(peaks)]
ube_d = np.unique(np.round(be_d, 1))[:len(peaks)]

for t, d, ref in zip(ube_t, ube_d, peaks):
    assert abs(t - ref) < etol
    assert abs(t - d) < etol

# NH3

filename = 'NH3'

data = np.loadtxt(filename + '-td' + '.dat')
be_t = data[:, 0]
f_t = data[:, 1]

data = np.loadtxt(filename + '-dos' + '.dat')
be_d = data[:, 0]
f_d = data[:, 1]

if output:
    plt.subplot(3, 1, 3)
    plt.bar(be_d - 0.1,
            f_d / f_d.max(),
            color='k',
            width=0.2,
            label='DFT eigenvalue')
    plt.bar(be_t - 0.1,
            f_t / f_t.max(),
            color='r',
            width=0.2,
            label='lr-TDDFT method')

    data = np.loadtxt(filename + '.dat')
    be_e = data[:, 0]
    f_e = 2.1 * data[:, 1]

    plt.plot(be_e, f_e, '-b', label='Exp')
    leg = plt.legend()
    plt.text(6.0, 0.8, 'NH$_3$', fontsize=22)
    plt.xlabel('Binding energy [eV]', fontsize=20)
    # ylabel('Spectroscopic factor', fontsize=20)
    plt.axis((5, 45, 0, 2))
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=18)

    plt.savefig('PES_fig.png')
    # show()

etol = 1.0
peaks = np.array([10.8, 16.2])
ube_t = np.unique(np.round(be_t, 1))[:len(peaks)]
ube_d = np.unique(np.round(be_d, 1))[:len(peaks)]

for t, d, ref in zip(ube_t, ube_d, peaks):
    assert abs(t - ref) < etol
    assert abs(t - d) < etol



import pickle
import numpy as np
from ase.parallel import paropen


ecut_equal = np.array([[19.157, 18.639, 18.502],
                       [11.814, 11.165, 10.974]])
for i, ecut in enumerate([100, 200, 300]):
    fil = pickle.load(paropen(f'C-g0w0_k8_ecut{ecut}_results.pckl', 'rb'))
    assert abs(fil['qp'][0, 0, 1] - ecut_equal[0, i]) < 0.01
    assert abs(fil['qp'][0, 0, 0] - ecut_equal[1, i]) < 0.01

freq_equal = np.array([19.91, 19.80, 19.89, 19.92, 19.90, 19.91])
for j, omega2 in enumerate([1, 5, 10, 15, 20, 25]):
    fil = pickle.load(paropen('C_g0w0_domega0_0.02_omega2_%s_results.pckl'
                              % omega2, 'rb'))
    assert abs(fil['qp'][0, 0, 1] - freq_equal[j]) < 0.01



from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters
from gpaw.atom.basis import BasisMaker

symbol = 'Au'
args = parameters[symbol]  # Dictionary of default setup parameters
args['rcut'] = 2.6  # Set cutoff of augmentation sphere

generator = Generator(symbol, 'RPBE')
generator.N *= 2  # Increase grid resolution
generator.run(write_xml=False, **args)

bm = BasisMaker(generator, name='special', run=False)

# Create double-zeta RPBE basis where p orbital is considered a valence state
# (ordinary dzp basis would use a smaller p-type Gaussian for polarization)
# The list jvalues indicates which states should be included, and the
# ordering corresponds to the valence states in the setup.
basis = bm.generate(zetacount=2, polarizationcount=0,
                    energysplit=0.1, jvalues=[0, 1, 2],
                    rcutmax=12.0)

basis.write_xml()  # Dump to file 'Au.special.dz.basis'



import os

import pytest
from ase.lattice.hexagonal import Graphene

from gpaw import GPAW, PW, FermiDirac
from gpaw.response.df import DielectricFunction
from ase.units import Hartree


def test_response_graphene_refined_response(in_tmp_dir):
    system = Graphene(symbol='C',
                      latticeconstant={'a': 2.467710, 'c': 1.0},
                      size=(1, 1, 1))
    system.pbc = (1, 1, 0)
    system.center(axis=2, vacuum=4.0)

    nk = 6
    nkrefine = 5

    kpt_refine = {'center': [1. / 3, 1. / 3, 0.],
                  'size': [nkrefine, nkrefine, 1],
                  'reduce_symmetry': False,
                  'q': [1. / (nk * nkrefine), 0, 0]}

    if not os.path.exists('graphene.gpw'):
        calc = GPAW(mode=PW(ecut=400),
                    nbands=8,
                    xc='PBE',
                    kpts={'size': [nk, nk, 1], 'gamma': True},
                    experimental={'kpt_refine': kpt_refine},
                    occupations=FermiDirac(0.026))
        system.calc = calc
        system.get_potential_energy()
        calc.write('graphene.gpw', 'all')

    df = DielectricFunction('graphene.gpw', eta=25e-3, domega0=0.01)
    alpha0x_w, alphax_w = df.get_polarizability(
        q_c=[1 / (nk * nkrefine), 0, 0])
    omega_w = df.get_frequencies()
    analyticalalpha_w = 1j / (8 * omega_w[1:] / Hartree)

    # Just some hardcoded test for alpha at omega=0
    assert alphax_w[0].real == pytest.approx(6.69, abs=0.04)

    if 0:
        from matplotlib import pyplot as plt
        plt.plot(omega_w, alphax_w.real, label='GPAW real part')
        plt.plot(omega_w, alphax_w.imag, '--', label='GPAW imag part')
        plt.plot(omega_w[1:], analyticalalpha_w.imag,
                 '--', label='Analytical imag part')
        plt.xlabel(r'$\hbar\omega$ (eV)')
        plt.ylabel(r'Polarizability $\mathrm{\AA}$')
        plt.legend()
        plt.xlim(0, 1)
        plt.ylim(0, 100)
        plt.show()



# Creates: si_abs.png
import numpy as np
from math import pi
import matplotlib.pyplot as plt
from matplotlib import rc
rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
# rc('text', usetex=True)
rc('figure', figsize=(4.0, 4.0), dpi=800)

data = np.loadtxt('df.csv', delimiter=',')

# Set plotting range
xmin = 0.1
xmax = 10.0
inds_w = (data[:, 0] >= xmin) & (data[:, 0] <= xmax)

plt.plot(data[inds_w, 0], 4 * pi * data[inds_w, 4])
plt.xlabel('$\\omega / [eV]$')
plt.ylabel('$\\mathrm{Im} \\epsilon(\\omega)$')
plt.savefig('si_abs.png', bbox_inches='tight')



"""Module for building atomic structures"""

from math import sqrt
from ase import Atoms, Atom
from ase.visualize import view


def fcc100(symbol, a, layers, L):
    """Build an fcc(100) surface

    Parameters
    ----------
    symbol: string
        Chemical symbol ('H', 'Li', ...).
    a: float
        Lattice constant.
    layers: int
        Number of layers.
    L: float
        Height of unit cell.

    """

    # Distance between atoms:
    d = a / sqrt(2)

    # Distance between layers:
    z = a / 2.

    assert L > layers * z, 'Unit cell too small!'

    # Start with an empty Atoms object:
    atoms = Atoms(cell=(d, d, L),
                  pbc=(True, True, False))

    # Fill in the atoms:
    for n in range(layers):
        position = [d / 2 * (n % 2),
                    d / 2 * (n % 2),
                    n * z]
        atoms.append(Atom(symbol, position))

    atoms.center(axis=2)

    return atoms


if __name__ == '__main__':

    fcc = fcc100('Al', 4.0, 4, 15.0)
    view(fcc * (4, 4, 2))



from gpaw.fdtd.poisson_fdtd import QSFDTD
from gpaw.fdtd.polarizable_material import (PermittivityPlus,
                                            PolarizableMaterial,
                                            PolarizableSphere)
from gpaw.tddft import photoabsorption_spectrum
from gpaw.mpi import world
import numpy as np

# Nanosphere radius (Angstroms)
radius = 50.0

# Whole simulation cell (Angstroms)
large_cell = np.array([3 * radius, 3 * radius, 3 * radius])

# Permittivity of Gold
# J. Chem. Phys. 135, 084121 (2011); http://dx.doi.org/10.1063/1.3626549
gold = [[0.2350, 0.1551, 95.62],
        [0.4411, 0.1480, -12.55],
        [0.7603, 1.946, -40.89],
        [1.161, 1.396, 17.22],
        [2.946, 1.183, 15.76],
        [4.161, 1.964, 36.63],
        [5.747, 1.958, 22.55],
        [7.912, 1.361, 81.04]]

# Initialize classical material
classical_material = PolarizableMaterial()

# Classical nanosphere
classical_material.add_component(
    PolarizableSphere(center=0.5 * large_cell,
                      radius=radius,
                      permittivity=PermittivityPlus(data=gold)))

# Quasistatic FDTD
qsfdtd = QSFDTD(classical_material=classical_material,
                atoms=None,
                cells=large_cell,
                spacings=[8.0, 1.0],
                communicator=world,
                remove_moments=(4, 1))

# Run ground state
energy = qsfdtd.ground_state('gs.gpw', nbands=-1)

# Run time evolution
qsfdtd.time_propagation('gs.gpw',
                        time_step=10,
                        iterations=1000,
                        kick_strength=[0.001, 0.000, 0.000],
                        dipole_moment_file='dm.dat')
# Spectrum
photoabsorption_spectrum('dm.dat', 'spec.dat', width=0.0)



from ase.io import read
from gpaw import GPAW
from ase.optimize.bfgslinesearch import BFGSLineSearch

slab = read('CH4Au532.xyz')
slab.set_cell([[7.309254, 0., 0.], [4.872836, 7.509545, 0.], [0., 0., 20.]],
              scale_atoms=False)
slab.set_pbc((1, 1, 1))

calc = GPAW(h=0.18, kpts=(4, 4, 1))
slab.calc = calc

dyn = BFGSLineSearch(slab, trajectory='relax.traj')
dyn.run(fmax=0.02)



from gpaw.response.df import DielectricFunction

df = DielectricFunction('gs_MoS2.gpw',
                        ecut=100,
                        frequencies=(0.,),
                        nbands=50,
                        intraband=False,
                        hilbert=False,
                        eta=0.1)

alpha = df.get_polarizability(pbc=[True, True, False],
                              filename=None)[1][0].real
print('alpha = ', alpha, 'AA')



from myqueue.task import task


def create_tasks():
    return [
        task('ruslab.py@8:10h'),
        task('ruslab.py+H@8:10h'),
        task('ruslab.py+N@8:10h'),
        task('ruslab.py+O@16:15h'),
        task('molecules.py@8:20m'),
        task('results.py',
             deps=['ruslab.py', 'ruslab.py+H', 'ruslab.py+N',
                   'ruslab.py+O', 'molecules.py'])]



# Creates: hybrid.png
import numpy as np
import pylab as plt

# Plot spectrum with r=0nm and r=5nm
spec0 = np.loadtxt('spec.1.dat')  # AuNP
spec1 = np.loadtxt('spec.2.dat')  # Na2
spec2 = np.loadtxt('spec.3.dat')  # AuNP+Na2

plt.figure()
plt.plot(spec0[:, 0], spec0[:, 1], 'r', label='Au nanoparticle')
plt.plot(spec1[:, 0], spec1[:, 1], 'g', label='Na$_2$')
plt.plot(spec1[:, 0], spec1[:, 1] + spec0[:, 1], 'k:',
         label='Sum of Na$_2$ and Au nanoparticle')
plt.plot(spec2[:, 0], spec2[:, 1], 'b', label='Na$_2$ near Au nanoparticle')
plt.legend(loc=1)
plt.xlabel('Energy [eV]', fontsize=12)
plt.ylabel('Dipole strength [1/eV]', fontsize=12)
plt.xlim((0, 5.0))
plt.ylim((-1, 22.5))
plt.savefig('hybrid.png')



from gpaw.xc.rpa import RPACorrelation

# This calculation is too heavy to run as an exercise!!

rpa1 = RPACorrelation('si.rpa.isolated.gpw', txt='si.atom.rpa_output.txt')

E1_i = rpa1.calculate(ecut=400.0)



import pytest
import numpy as np

from ase import Atoms

from gpaw import GPAW, FermiDirac, Davidson
from gpaw.wavefunctions.pw import PW
from gpaw.response.chi0 import Chi0

# This script asserts that the chi's obtained
# from GS calculations using symmetries
# and GS calculations not using symmetries return
# the same results. Tests that the chi's are element-wise
# equal to a tolerance of 1e-10.


@pytest.mark.slow
def test_response_symmetry(in_tmp_dir):
    pwcutoff = 400.0
    k = 4
    a = 4.59
    c = 2.96
    u = 0.305

    rutile_cell = [[a, 0, 0],
                   [0, a, 0],
                   [0, 0, c]]

    TiO2_basis = np.array([[0.0, 0.0, 0.0],
                           [0.5, 0.5, 0.5],
                           [u, u, 0.0],
                           [-u, -u, 0.0],
                           [0.5 + u, 0.5 - u, 0.5],
                           [0.5 - u, 0.5 + u, 0.5]])

    bulk_crystal = Atoms(symbols=['Ti', 'Ti', 'O', 'O', 'O', 'O'],
                         scaled_positions=TiO2_basis,
                         cell=rutile_cell,
                         pbc=(1, 1, 1))
    data_s = []
    for symmetry in ['off', {}]:
        bulk_calc = GPAW(mode=PW(pwcutoff),
                         nbands=42,
                         eigensolver=Davidson(1),
                         kpts={'size': (k, k, k), 'gamma': True},
                         xc='PBE',
                         occupations=FermiDirac(0.00001),
                         parallel={'band': 1},
                         symmetry=symmetry)

        bulk_crystal.calc = bulk_calc
        _ = bulk_crystal.get_potential_energy()
        bulk_calc.write('bulk.gpw', mode='all')
        X = Chi0('bulk.gpw')
        chi_t = X.calculate([1. / 4, 0, 0])[1:]
        data_s.append(list(chi_t))

    msg = 'Difference in Chi when turning off symmetries!'

    while len(data_s):
        data1 = data_s.pop()
        for data2 in data_s:
            for dat1, dat2 in zip(data1, data2):
                if dat1 is not None:
                    assert (np.abs(dat1 - dat2).max() ==
                            pytest.approx(0, abs=0.001)), msg



import pytest
import numpy as np
from ase.parallel import parprint
from ase.lattice.compounds import L1_2
from gpaw import GPAW, PW, Mixer, Davidson


@pytest.mark.slow
def test_xc_qna_stress(in_tmp_dir):
    name = 'Cu3Au'
    # structure = 'L1_2'
    ecut = 300
    kpts = (2, 2, 2)

    QNA = {'alpha': 2.0,
           'name': 'QNA',
           'stencil': 1,
           'orbital_dependent': False,
           'parameters': {'Au': (0.125, 0.1), 'Cu': (0.0795, 0.005)},
           'setup_name': 'PBE',
           'type': 'qna-gga'}

    atoms = L1_2(['Au', 'Cu'], latticeconstant=3.7)

    calc = GPAW(mode=PW(ecut),
                eigensolver=Davidson(2),
                nbands='120%',
                mixer=Mixer(0.4, 7, 50.0),
                parallel=dict(domain=1),
                xc=QNA,
                kpts=kpts,
                txt=name + '.txt')

    atoms.calc = calc

    atoms.set_cell(np.dot(atoms.cell,
                          [[1.02, 0, 0.03],
                           [0, 0.99, -0.02],
                           [0.2, -0.01, 1.03]]),
                   scale_atoms=True)

    s_analytical = atoms.get_stress()
    s_numerical = atoms.calc.calculate_numerical_stress(atoms, 1e-5)
    s_err = s_numerical - s_analytical

    parprint('Analytical stress:', s_analytical)
    parprint('Numerical stress :', s_numerical)
    parprint('Error in stress  :', s_err)
    assert np.all(abs(s_err) < 5e-4)



from ase.io import read
from ase.units import mol, kcal
# calculate solvation Gibbs energy in various units
Egasphase = read('gasphase.txt').get_potential_energy()
Ewater = read('water.txt').get_potential_energy()
DGSol_eV = Ewater - Egasphase
DGSol_kcal_per_mol = DGSol_eV / (kcal / mol)
assert abs(DGSol_kcal_per_mol - -4.5) < 0.05



import numpy as np
from ase import Atoms
from gpaw import GPAW

d = 1.54
dx = d * (2 / 3)**0.5
dz = d / 3**0.5
h = 1.1


def cnh2n(n):
    assert n % 2 == 1
    positions = []
    for i in range(n):
        x = i * dx
        z = dz * (i % 2)
        positions.append((x, 0, z))
        if i == 0:
            positions.append((x, 0, -h))
            positions.append((x - dx, 0, dz))
        elif i == n - 1:
            positions.append((x, 0, -h))
            positions.append((x + dx, 0, dz))
        elif i % 2 == 0:
            positions.append((x, -h * (2 / 3)**0.5, -h / 3**0.5))
            positions.append((x, h * (2 / 3)**0.5, -h / 3**0.5))
        else:
            positions.append((x, -h * (2 / 3)**0.5, z + h / 3**0.5))
            positions.append((x, h * (2 / 3)**0.5, z + h / 3**0.5))

    atoms = Atoms(f'(CH2){n}', positions)
    atoms.set_distance(0, 2, h, 0)
    atoms.set_distance(-3, -1, h, 0)
    magmoms = np.zeros(3 * n)
    magmoms[0] = 1.0
    magmoms[-3] = 1.0
    atoms.set_initial_magnetic_moments(magmoms)
    return atoms


if __name__ == '__main__':
    for n in range(3, 13, 2):
        atoms = cnh2n(n)
        d = atoms.get_distance(0, -3)
        atoms.center(vacuum=2 * d)
        name = f'C{n}H{2 * n}'
        atoms.calc = GPAW(mode='lcao',
                          basis='dzp',
                          xc='PBE',
                          txt=name + '.txt')
        atoms.get_potential_energy()
        atoms.calc.write(name + '.gpw', 'all')



if 1:  # example: set eigensolver/mixer
    from gpaw.eigensolvers.davidson import Davidson
    from gpaw import Mixer, MixerSum

    def calc(atoms):
        if atoms.get_initial_magnetic_moments().any():
            M = MixerSum
        else:
            M = Mixer
        atoms.calc.set(eigensolver=Davidson(), mixer=M(0.05, 5, 50))



from ase import Atoms
from ase.io import write
from ase.visualize import view
from gpaw import restart
from gpaw.wannier import calculate_overlaps

atoms, calc = restart('benzene.gpw')

homo = calc.get_pseudo_wave_function(band=14)
write('homo.cube', atoms, data=homo)

# Initialize the Wannier class
w = calculate_overlaps(calc, nwannier=15).localize_er()
view(atoms + Atoms(symbols='X15', positions=w.centers))

# Find the index of the center with the lowest y-coordinate:
nsigma = w.centers[:, 1].argmin()
sigma = w.get_function(calc, nsigma)

write('benzene.xyz', atoms)
write('sigma.cube', atoms, data=sigma)



# Creates: gaps.csv
from ase.io import read


fd = open('gaps.csv', 'w')
gaps = []
for name in ['no_u', 'normalized_u', 'not_normalized_u']:
    n = read(name + '.txt')
    gap = n.calc.get_eigenvalues(spin=1)[1] - n.calc.get_eigenvalues(spin=0)[1]
    gaps.append(gap)
    print(f"{name.replace('_', ' ').replace('u', 'U')}, {gap:.3f}",
          file=fd)

assert abs(gaps[1] - gaps[0] - 6.0) < 0.8



from ase.build import mx2
from gpaw import GPAW, PW, FermiDirac

calc = GPAW(mode=PW(600),
            xc='PBE',
            occupations=FermiDirac(width=0.001),
            kpts={'size': (6, 6, 1), 'gamma': True},
            txt='gs_Sn.txt')

slab = mx2(formula='SnSn2', a=4.67, thickness=1.7, vacuum=5.0)
del slab[-1]
slab.calc = calc
slab.get_potential_energy()

calc.write('gs_Sn.gpw')



# creates: paw_note.pdf
import os
import shutil
import subprocess
from pathlib import Path

import matplotlib.pyplot as plt

from gpaw.atom.aeatom import AllElectronAtom

ae = AllElectronAtom('Pt')
ae.run()
ae.plot_wave_functions(show=False)
rcut = 2.5
lim = [0, 3.5, -2, 4]
plt.plot([rcut, rcut], lim[2:], 'k--', label='_nolegend_')
plt.axis(lim)
plt.text(0.6, 2, '[Pt] = [Xe]4f$^{14}$5d$^9$6s$^1$')
plt.savefig('Pt.png', dpi=80)

dir = os.environ.get('PDF_FILE_DIR')
if dir:
    shutil.copyfile(Path(dir) / 'paw_note.pdf', 'paw_note.pdf')
else:
    try:
        subprocess.run(
            'pdflatex -interaction=nonstopmode paw_note > /dev/null && '
            'bibtex paw_note > /dev/null && '
            'pdflatex -interaction=nonstopmode paw_note > /dev/null && '
            'pdflatex -interaction=nonstopmode paw_note > /dev/null',
            shell=True, check=True)
    except subprocess.CalledProcessError:
        subprocess.run('echo "No pdflatex" > paw_note.pdf', shell=True)

subprocess.run(['cp', 'paw_note.pdf', '..'])



import numpy as np
import pytest

from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters, tf_parameters
from gpaw import setup_paths
from gpaw import mpi


def equal(x, y, tolerance=0):
    assert x == pytest.approx(y, abs=tolerance)


def print_reference(data_i, name='ref_i', fmt='%.12le'):
    if mpi.world.rank == 0:
        print('%s = [' % name, end='')
        for i, val in enumerate(data_i):
            if i > 0:
                print('', end='\n')
                print(' ' * (len(name) + 4), end='')
            print(fmt % val, end='')
            print(',', end='')
        print('\b]')


def findpeak(x, y):
    dx = x[1] - x[0]
    i = y.argmax()
    a, b, c = np.polyfit([-1, 0, 1], y[i - 1:i + 2], 2)
    assert a < 0
    x = -0.5 * b / a
    return dx * (i + x), a * x**2 + b * x + c


def gen(symbol, exx=False, name=None, yukawa_gamma=None,
        write_xml=False, **kwargs):
    setup = None
    if mpi.rank == 0:
        if 'scalarrel' not in kwargs:
            kwargs['scalarrel'] = True
        g = Generator(symbol, **kwargs)
        if 'orbital_free' in kwargs:
            setup = g.run(exx=exx, name=name, yukawa_gamma=yukawa_gamma,
                          use_restart_file=False,
                          write_xml=write_xml,
                          **tf_parameters.get(symbol, {'rcut': 0.9}))
        else:
            setup = g.run(exx=exx, name=name, yukawa_gamma=yukawa_gamma,
                          use_restart_file=False,
                          write_xml=write_xml,
                          **parameters[symbol])
    setup = mpi.broadcast(setup, 0)
    if setup_paths[0] != '.':
        setup_paths.insert(0, '.')
    return setup



"""A proton in an electric field."""
from ase import Atoms
from ase.units import Hartree, Bohr

from gpaw import GPAW
from gpaw.external import ConstantElectricField


def test_ext_potential_constant_e_field(in_tmp_dir):
    h = Atoms('H')
    h.center(vacuum=2.5)
    h.calc = GPAW(external=ConstantElectricField(1.0),  # 1 V / Ang
                  charge=1,
                  txt='h.txt')
    e = h.get_potential_energy()
    f1 = h.get_forces()[0, 2]
    h[0].z += 0.001
    de = h.get_potential_energy() - e
    f2 = -de / 0.001
    print(f1, f2)
    assert abs(f1 - 1) < 1e-4
    assert abs(f2 - 1) < 5e-3

    # Check writing and reading:
    h.calc.write('h')
    vext = GPAW('h', txt=None).hamiltonian.vext
    assert abs(vext.field_v[2] - 1.0 * Bohr / Hartree) < 1e-13



import pytest
from gpaw.solvation.sjm import SJM, SJMPower12Potential

from ase.build import fcc111
from gpaw import FermiDirac

# Import solvation modules
from ase.data.vdw import vdw_radii
from gpaw.solvation import (
    EffectivePotentialCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction)


@pytest.mark.skip(reason='TODO')
def test_sjm():
    # Solvent parameters
    u0 = 0.180  # eV
    epsinf = 78.36  # Dielectric constant of water at 298 K
    gamma = 0.00114843767916  # 18.4*1e-3 * Pascal* m
    T = 298.15   # K

    def atomic_radii(atoms):
        return [vdw_radii[n] for n in atoms.numbers]

    # Structure is created
    atoms = fcc111('Au', size=(1, 1, 3))
    atoms.center(axis=2, vacuum=8)
    atoms.translate([0, 0, -2])

    # SJM parameters
    potential = 3.0
    ne = 1.20
    dpot = 0.05

    # The calculator
    calc = SJM(doublelayer={'upper_limit': 19.5},
               potential=potential,
               dpot=dpot,
               ne=ne,
               gpts=(8, 8, 64),
               poissonsolver={'dipolelayer': 'xy'},
               kpts=(1, 1, 1),
               xc='PBE',
               occupations=FermiDirac(0.1),
               cavity=EffectivePotentialCavity(
                   effective_potential=SJMPower12Potential(atomic_radii, u0),
                   temperature=T,
                   surface_calculator=GradientSurface()),
               dielectric=LinearDielectric(epsinf=epsinf),
               interactions=[SurfaceInteraction(surface_tension=gamma)])

    # Run the calculation
    atoms.calc = calc
    atoms.get_potential_energy()
    assert abs(calc.get_electrode_potential() - potential) < dpot
    elpot = calc.get_electrostatic_potential().mean(0).mean(0)
    assert abs(elpot[2] - elpot[10]) < 1e-3



import os
import sys
import pylab
n = 1


def show():
    global n
    pylab.savefig('x%d.png' % n)
    n += 1


pylab.show = show

if 0:
    def view(*args, **kwargs):
        return
    import ase
    ase.view = view

os.system('rm -rf test; mkdir test')
os.chdir('test')
for dir, script in [
    ('neb', 'neb1.py'),
    ('aluminium', 'Al_fcc.py'),
    ('aluminium', 'Al_fcc_convergence.py'),
    ('surface', 'Al100.py'),
    ('surface', 'work_function.py'),
    ('diffusion', 'initial.py'),
    ('diffusion', 'solution.py'),
    ('diffusion', 'densitydiff.py'),
    ('vibrations', 'h2o.py'),
    ('vibrations', 'H2O_vib.py'),
    ('iron', 'ferro.py'),
    ('iron', 'anti.py'),
    ('iron', 'non.py'),
    ('iron', 'PBE.py'),
    ('band_structure', 'Na_band.py'),
    ('band_structure', 'plot_band.py'),
    ('wannier', 'si.py'),
    ('wannier', 'wannier-si.py'),
    ('wannier', 'benzene.py'),
    ('wannier', 'wannier-benzene.py'),
    ('stm', 'HAl100.py'),
    ('wavefunctions', 'CO.py'),
    ('dos', 'pdos.py'),
    ('lrtddft', 'ground_state.py'),
    ('transport', 'pt_h2_tb_transport.py'),
    ('transport', 'pt_h2_lcao.py'),
    ('transport', 'pt_h2_lcao_transport.py')]:
    exec(open('../' + dir + '/' + script).read(), {'k': 6, 'N': 5})
for dir, script, args in [
    ('stm', 'stm.py', ['HAl100.gpw']),
    ('dos', 'dos.py', ['Al-fcc.gpw', 'si.gpw', 'CO.gpw',
                       'ferro.gpw', 'anti.gpw', 'non.gpw'])]:
    for arg in args:
        sys.argv = ['', arg]
        exec(open('../' + dir + '/' + script).read())



from pathlib import Path
from typing import Union

import numpy as np


def read_bader_charges(filename: Union[str, Path] = 'ACF.dat') -> np.ndarray:
    path = Path(filename)
    charges = []
    with path.open() as fd:
        for line in fd:
            words = line.split()
            if len(words) == 7:
                charges.append(float(words[4]))
    return np.array(charges)


if __name__ == '__main__':
    import subprocess
    import sys
    from ase.io import write
    from ase.units import Bohr
    from gpaw import GPAW
    from gpaw.utilities.ps2ae import PS2AE

    calc = GPAW(sys.argv[1])
    converter = PS2AE(calc, grid_spacing=0.05)
    density = converter.get_pseudo_density()
    ne = density.sum() * converter.dv
    print(ne, 'electrons')
    write('density.cube', calc.atoms, data=density * Bohr**3)
    subprocess.run('bader -p all_atom density.cube'.split())
    charges = read_bader_charges()
    for setup, charge in zip(calc.density.setups, charges):
        charge -= setup.Nv
        print(f'{setup.symbol:2} {charge:10.6f}')



from ase.build import bulk
from gpaw import GPAW, PW, FermiDirac

calc = GPAW(mode=PW(600),
            xc='PBE',
            occupations=FermiDirac(0.01),
            kpts=(32, 32, 16),
            txt='gs_Co.txt')

bulk = bulk('Co', 'hcp')
bulk.set_initial_magnetic_moments([1.0, 1.0])
bulk.calc = calc
bulk.get_potential_energy()

calc.write('gs_Co.gpw')



# Creates: xas_h2o_spectrum.png, h2o_xas_box.png


def create_tasks():
    from myqueue.task import task
    return [
        task('setups.py'),
        task('run.py@8:25m', deps='setups.py'),
        task('dks.py@8:25m', deps='setups.py'),
        task('h2o_xas_box1.py@8:25m', deps='setups.py'),
        task('submit.agts.py', deps='run.py,dks.py,h2o_xas_box1.py')]


if __name__ == '__main__':
    exec(open('plot.py').read())
    e_dks = float(open('dks.result').readline().split()[2])
    assert abs(e_dks - 532.502) < 0.001
    exec(open('h2o_xas_box2.py').read())



"""Test selfconsistent RSF calculation with Yukawa potential including vc."""
import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW, setup_paths, KohnShamConvergenceError
from gpaw.xc.hybrid import HybridXC
from gpaw.poisson import PoissonSolver
from gpaw.occupations import FermiDirac
from gpaw.test import gen
from gpaw.eigensolvers import RMMDIIS
from gpaw.cluster import Cluster

pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


def test_exx_exx_scf(in_tmp_dir):
    if setup_paths[0] != '.':
        setup_paths.insert(0, '.')

    h = 0.3

    # No energies - simpley convergence test, esp. for 3d TM

    # for atom in ['F', 'Cl', 'Br', 'Cu', 'Ag']:
    for atom in ['Ti']:
        gen(atom, xcname='PBE', scalarrel=False, exx=True)
        work_atom = Cluster(Atoms(atom, [(0, 0, 0)]))
        work_atom.minimal_box(4, h=h)
        work_atom.translate([0.01, 0.02, 0.03])
        work_atom.set_initial_magnetic_moments([2.0])
        calculator = GPAW(convergence={'energy': 0.01,
                                       'eigenstates': 3,
                                       'density': 3},
                          eigensolver=RMMDIIS(),
                          poissonsolver=PoissonSolver(use_charge_center=True),
                          occupations=FermiDirac(width=0.0, fixmagmom=True),
                          h=h, maxiter=35)   # Up to 24 are needed by now
        calculator.set(xc=HybridXC('PBE0'))
        calculator.set(txt=atom + '-PBE0.txt')
        work_atom.calc = calculator
        try:
            work_atom.get_potential_energy()
        except KohnShamConvergenceError:
            pass
        assert calculator.scf.converged, 'Calculation not converged'
